import sys
import pandas as pd
import re
from yapsy.IPlugin import IPlugin


# modify search path to include parent directory
sys.path.append("../")

import sccm
import gbls


class PluginOne(IPlugin):
    def print_name(self):
        print (
            'Plugin 1 for custom Vulnmine Input'
            ' has loaded successfully.'
            )


    def modify_hosts(self, my_hosts):
        """
        Functions
        =========

        _classify_using_sccm_data
            Classify discovered hosts using SCCM discovery data.
        _classify_using_DN_OUs
            Classify hosts by membership in AD OUs
        _classify_using_ad_grps
            Use membership in AD groups to classify hosts.

        """
        # Updated dframe
        new_df_sys = ''

        def _classify_using_sccm_data():
            """Classify discovered hosts using sccm v_R_System data.

            Actions
            =======

            To categorize by Region:

                The AD_Site_Name0 is used to map hosts to a Region.

            Hardcoded lists of sites are then used to classify by region.

            """
            ######
            #   Classify Sites by Region based on AD_Site_Name0 field in
            #       v_R_System record
            ######

            REGION_A = [u'NTH', u'WST']
            REGION_B = [u'STH', u'EST']

            print(
                '\n\n\nEntering plugin1 - _classify_using_sccm_data\n\n')

            # Map Site to corresponding Region

            def __classify_site_by_region(mysite):

                try:
                    if mysite in REGION_A:
                        myregion = 'Region_A'
                    elif mysite in REGION_B:
                        myregion = 'Region_B'
                    else:
                        myregion = 'Unknown'

                except:
                    print(
                        '\n\n***Error in classifying by '
                        '\nregion for value: "{0}"\n\n'.format(
                            mysite))
                    myregion = None
                return myregion

            # produce series with Region for each site
            s_myRegion = new_df_sys['Site_X'].apply(
                __classify_site_by_region
                ).astype('category')

            # place into the main dataframe
            new_df_sys['Region_X'] = pd.Series(
                                        s_myRegion,
                                        index=new_df_sys.index)

            # print() output Region values
            print(
                '\n\nRegions: \n{0}\n\n'.format(
                        pd.unique(new_df_sys['Region_X'].values)
                        )
                )

            # Number of hosts in each region
            print(
                '\n# hosts in each region: \n{0}\n\n'.format(
                    new_df_sys['Region_X'].value_counts()
                    )
                )
            return None





        def _classify_using_DN_OUs():
            """Classify hosts member of an AD OU.

            Actions
            =======
            Extract various patterns from 'Distinguished_Name0' field of
            the v_R_System table. Use these values to classify the hosts.

            Exceptions
            ==========
            IOError     Produce error message and then ignore

            """
            print('\n\nEntering plugin1 - _classify_using_DN_OUs\n\n')

            # Extract the Distinguished_Name0 field as a series and
            # prepare for pattern matching
            s_dn = new_df_sys[
                        'Distinguished_Name0'].str.strip().str.lower()

            # Look for generic desktops/laptops/servers

            pattern = re.compile(
                r'ou=[0-9A-Za-z_ ]*(desktop|laptop|server)',
                re.IGNORECASE | re.UNICODE
                )

            new_df_sys['HostFn_X'] = s_dn.str.extract(
                                            pattern,
                                            expand=False).astype('category')

            print(
                '\n\n# hosts of each type:\n{0}'.format(
                                    new_df_sys['HostFn_X'].value_counts()
                                    )
                )
            return None


        def _classify_using_ad_grps():
            """Classify hosts member of an AD group.

            Actions
            =======
            Read CSV format data which lists contents of the AD groups.

            Mark hosts that are members of these groups.

            Exceptions
            ==========
            IOError     Produce error message and then ignore

            """
            print('\n\nEntering plugin1 - _classify_using_ad_grps\n\n')

            try:
                df_ad_vip = pd.io.parsers.read_csv(
                                    gbls.ad_vip_grps,
                                    sep=gbls.SEP2,
                                    error_bad_lines=False,
                                    warn_bad_lines=True,
                                    quotechar='"',
                                    comment=gbls.HASH,
                                    encoding='utf-16')

            except IOError as e:
                print('\n\n***I/O error({0}): {1}\n\n'.format(
                            e.errno, e.strerror))

            # ValueError Exception could mean empty data set read
            # Initialize an empty dataframe
            except ValueError as e:
                print(
                    '\n\n***Value error: {0}\n- empty data set '
                    'returned\n\n'.format(
                                sys.exc_info()[0]
                                )
                    )
                df_ad_vip = pd.DataFrame({'distinguishedName': []})

            except:
                print(
                    '\n\n***Unexpected error: {0}\n\n'.format(
                        sys.exc_info()[0])
                        )
                raise

            # print() basic information
            print(
                '\n\nRaw input: Hosts in VIP AD '
                'group : \n{0}\n{1}\n\n'.format(
                    df_ad_vip.shape,
                    df_ad_vip.columns)
                )

            # mark hosts that are in the VIP AD group
            crit_ad_excl = new_df_sys.Distinguished_Name0.isin(
                df_ad_vip['distinguishedName'])
            new_df_sys.loc[crit_ad_excl, 'VIP_X'] = 'vip'

            # count of marked hosts
            print(
                '\n\nSCCM-managed hosts in '
                'VIP AD group: \n{0}\n\n'.format(
                        new_df_sys['VIP_X'].value_counts()
                        )
                )

            print(
                '\nSystem dataframe with '
                'additional classification columns: \n{0}\n{1}\n\n'.format(
                        new_df_sys.shape,
                        new_df_sys.columns
                        )
                )
            return None

        #####
        #   Mainline code
        #####
        # Read packed hosts dframe
        my_hosts.load()
        new_df_sys = my_hosts.get()

        print (
            '\nPlugin I/P sccm dframe:\n{0}\n{1}'.format(
                                                new_df_sys.shape,
                                                new_df_sys.columns
                                                )
            )

        # Classify hosts
        _classify_using_sccm_data()
        _classify_using_DN_OUs()
        _classify_using_ad_grps()

        # Modify hosts object internal dframe directly.

        # This is more direct and probably safer than pandas to_csv /
        # read_csv

        my_hosts.df_sys = new_df_sys
        my_hosts.save()
        return None


"""ml: Do Machine Learning Classification of test data.

Purpose
=======

The ml module does tha ML Classification using the Random Forest
Classifier.

There are in fact 2 different models being used:
*   CPE "Vendor" to SCCM "Publisher" matching
*   CPE "Software" to SCCM "Software" inventory data

The appropriate model is chosen at class initialization.

Manually labelled data is used to update the data to be classified. This
eliminates "known" matches from further processing.

The actual ML classification is done.

Then data is post-processed. Any duplicate records are eliminated. Then the
manually matched data is appended to the new classified data to form the final
dataframe.

Public classes
==============

MLClassify      ML Classification using Random Forest Classifier


Restrictions
------------
The following restrictions should be clearly understood and respected.

*  Fields and features *have* to be exactly as specified (same order, same
   names) as in the initialization code.

*  Labelled data always has a "match" attribute field.
"""

import pandas as pd
import numpy as np

from sklearn.externals import joblib

import logging
import sys

import gbls
import utils

# Public classes
__all__ = (
        'MLClassify'
        )


class MLClassify(object):
    """Match NVD CPE "Software" data to SCCM "Software" inventory data.

    Actions
    -------

    *  There are two different models being used for classification. The
       appropriate model is input when the class is initialized. At the same
       time the key and feature lists are initializedd acccordingg to the
       model chosen.

    *  The new data to be classified is updated using manually labelled data.
       This ensures that only new unclassified data is input to ML
       classification.

    *  The Random Forest Classification algorithm is run.

    *  The newly-classifed test data is concatenated with the manually
       labelled data. Duplicates are eliminated to form the final resulting
       dataframe.

    Methods
    -------
    __init__    The class constructor which:
                - initializes logging
                - initializes the key list and feature list variables
                - reads in the ML Model serialized data.

    upd_using_labelled_data
                Updates data to be classified using the manually labelled data

    ml_classify Do the ML Random Forest classification

    post_process_matched_data
                Concatenate the new classified data with the manally labelled
                data. Eliminate any duplicate data records.


    Exceptions
    ----------
    IOError     Log error message and ignore

    Restrictions
    ------------

    The key and feature field lists *must* be kept in the specified order
    since that is # how the ML algorithm was originally trained.

    (See the sccmgbl module for the corresponding definitions.)

    Returns
    -------
    None

    """

    def __init__(
            self,
            type_data=None,
            mylogger=None
            ):
        """Initialize class by configuring logging,  initializing dataframe.

        Actions
        -------

        This is the class constructor. To initialize:

        * Initialize logging
        * Initialize class attributes: key_list, feature_list, attr_list
        * Input the correct model from the serialized disk file.

        I/P Parameters
        --------------
        type_data   specifies which model is to be used:
                        'vendor'    CPE Vendor - SCCM Publisher0
                        'software'  CPE Sofware - SCCM Software inventory

        mylogger    logging object. If None, then a new object is initialzed.

        """
        # Configure logging

        if mylogger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(gbls.loglvl)
        else:
            self.logger = mylogger

        # Initialize key list and feature list for this model
        # Input the respective serialized ML model

        self.logger.info(
                '\n\nInitializing ML_Match class\n\n'
                'Type of data to be classified: \n{0}\n\n'.format(type_data)
                )

        if (type_data == 'vendor'):

            self._key_list = gbls.vendor_key_list
            self._feature_list = gbls.vendor_feature_list
            self._attr_list = gbls.vendor_attr_list

            model = gbls.clf_vendor

        elif (type_data == 'software'):

            self._key_list = gbls.sft_key_list
            self._feature_list = gbls.sft_feature_list
            self._attr_list = gbls.sft_attr_list

            model = gbls.clf_software

        else:
            self.logger.critical(
                '*** Bad input when initializing the ML_Match class\n\n'
                )
            return None

        # Input the serialized ML model

        try:
            self.__clf = joblib.load(model)

        except IOError as e:
            self.logger.critical(
                    '*** I/O error ML Model({0}): {1}\n\n'.format(
                                                e.errno, e.strerror
                                                )
                    )

        except:
            self.logger.critical(
                '*** Unexpected error loading ML model: {0}\n\n'.format(
                    sys.exc_info()[0]))
            raise

    def upd_using_labelled_data(self, p_df_match, p_df_match_lbl):
        """Update the data to be matched with the manually labelled data.

        Actions
        -------

        *   Discard columns that are not in the key_list. Drop records with
            null fields.

        *   Update the data to be matched (i.e. the cartesian product data)
            with "match" values from the labelled data.

        Returns
        -------
        Updated dataframe

        """
        self.logger.info('\n\nEntering upd_using_labelled_data\n\n')

        # force call-by-value
        df_match = p_df_match.copy()
        df_match_lbl = p_df_match_lbl.copy()

        if df_match_lbl.empty:
            self.logger.info('Input dataframe df_match_lbl is empty.')
            return (df_match)

        if df_match.empty:
            self.logger.info('Input dataframe df_match is empty.')
            return (df_match)

        self.logger.info(
                'Data set to be classified: {0}\n{1}\n{2}\n\n'
                'Labelled data: \n{3}\n{4}\n{5}\n\n'.format(
                                    df_match.shape,
                                    df_match.columns,
                                    df_match.apply(pd.Series.nunique),
                                    df_match_lbl.shape,
                                    df_match_lbl.columns,
                                    df_match_lbl.apply(pd.Series.nunique)
                                    )
                )

        # Update the data to be matched with the results from manual
        # labelling effort. After doing all this hard manual
        # classification effort, why not use it? df_match_upd =
        # pd.merge(df_match, df_match_lbl, how='left', on=['publisher0',
        # 'vendor_X'])

        # First keep only relevant columns from labelled data
        df_match_lbl1 = df_match_lbl.loc[
                                        :,
                                        self._key_list + ['match']
                                        ]

        # Next update new input data with known values from
        # the labelled data
        df_match_upd0 = pd.merge(
                            df_match,
                            df_match_lbl1,
                            how='left',
                            on=self._key_list
                            )

        # drop records with null fields (if any)

        df_match_upd = df_match_upd0.dropna(
                                how='any',
                                subset=self._attr_list
                                )

        # check that only Match values changed
        self.logger.debug(
                    '\n\nCheck that updating with the labelled '
                    'data did not add extra null records.\n'
                    '--Updated data set to be classified:'
                    '\n{0}\n{1}\n{2}\n\n'.format(
                            df_match_upd.shape,
                            df_match_upd.columns,
                            df_match_upd.apply(pd.Series.nunique)
                            )
            )

        return (df_match_upd)

    def ml_classify(self, p_df_match_upd):
        """Do the Machine Learning Classification.

        Actions
        -------

        *   Separate the test data to be classified from the data which has
            already been labelled.

        *   Format the test data as a numpy array and run the Random Forest
            Classification algorithm.

        *   Update the I/P test dataframe with the classification match
            results.

        Returns
        -------
        Dataframe containing classified test data
        Dataframe containing manually labelled data

        """
        # Force call by value
        df_match_upd = p_df_match_upd.copy()

        # Do ML classification
        self.logger.info('\n\nEntering ml_classify\n\n')

        if df_match_upd.empty:
            self.logger.critical(
                '*** ML: Input dframe is empty!'
                )
            return(df_match_upd, df_match_upd)

        # Separate out the test data (i.e. not yet classified)
        df_match_test = df_match_upd[df_match_upd['match'].isnull()]
        df_match_labelled = df_match_upd[
                                df_match_upd['match'].notnull()
                                ]

        if df_match_test.empty:
            df_match_test2 = df_match_test.reset_index(drop=True)
            self.logger.info(
                '\n\n ML: No elements to classify!'
                )
        else:
            self.logger.info(
                '\nStarting ML matching\n\n'
                )

            # Format the test data feature set.
            df_match_test1 = df_match_test[['match'] + self._feature_list]

            # Convert to a numpy array for input to the ML algorithm
            np_match_test1 = np.asarray(df_match_test1)
            Xt = np_match_test1[:, 1:]

            s_match_test = pd.Series(self.__clf.predict(Xt))
            df_match_test2 = df_match_test.reset_index(drop=True)
            df_match_test2['match'] = s_match_test

            # Most, if not all, test data pairs will be rejected
            # since the labelling effort was quite comprehensive

            self.logger.info(
                '\n\nResults of ML '
                'classification: \n\n'
                'Test data: {0}\n{1}\n '
                'Labelled data: {2}\n{3}\n '
                'Match counts: {4}\n'.format(
                    df_match_test2.shape,
                    df_match_test2.columns,
                    df_match_labelled.shape,
                    df_match_labelled.columns,
                    df_match_test2['match'].value_counts()
                    )
                )

            sample_size = min(
                            (df_match_test2.match == 1).sum(),
                            10
                            )

            if sample_size > 0:
                self.logger.info(
                        '\nSample matches: \n{0}\n\n'.format(
                        df_match_test2[
                                df_match_test2.match == 1
                                ].sample(sample_size)
                        )
                    )
            else:
                self.logger.info(
                        '\nNo matches!'
                        )

        return (df_match_test2, df_match_labelled)

    def post_process_matched_data(
                self,
                p_df_match_test2,
                p_df_match_lbl
                ):
        """Post-process the matched and labelled data.

        Actions
        -------

            Concatenate the manually labelled data with the data freshly
            classified by the ML algorithm.

            Eliminate duplicate records, if any.

            Keep +ve matches only.

        Returns
        -------

        Dataframe containing consolidated match data: ML-classified test data
        and labelled data

        """
        # Force call-by-value
        df_match_test2 = p_df_match_test2.copy()
        df_match_lbl = p_df_match_lbl.copy()

        # concatenate the two sets of classified data: the manual set, and
        # the machine-classified one

        self.logger.info('\n\nEntering post_process_matched_data\n\n')

        if df_match_lbl.empty:
            self.logger.info(
                'Input dataframe df_match_labelled is empty.\n\n'
                )
            return (df_match_test2)

        if df_match_test2.empty:
            self.logger.info(
                'Input dataframe df_match_test2 is empty.\n\n'
                )
            return (df_match_lbl)

        self.logger.info(
            'ML-matched data: {0}\n{1}\n{2}\n'
            'Labelled data: {3}\n{4}\n{5}\n\n'.format(
                                df_match_test2.shape,
                                df_match_test2.columns,
                                df_match_test2.apply(pd.Series.nunique),
                                df_match_lbl.shape,
                                df_match_lbl.columns,
                                df_match_lbl.apply(pd.Series.nunique)
                                )
            )

        df_match_consol1 = pd.concat([
                                df_match_test2[
                                    df_match_test2['match'].notnull()
                                    ],
                                df_match_lbl
                                ],
                                ignore_index=True
                                )

        # eliminate any possible remaining duplicate records
        df_match_consol = df_match_consol1.drop_duplicates(self._key_list)

        self.logger.info(
            '\nConsolidated Matched dataframe: '
            '\n{0}\n{1}\n{2}\n{3}\n\n'.format(
                    df_match_consol.shape,
                    df_match_consol.columns,
                    df_match_consol['match'].value_counts(),
                    df_match_consol.apply(pd.Series.nunique)
                    )
            )

        # Only interested in +ve matches
        df_match_consol1 = df_match_consol[
                                df_match_consol['match'] == 1
                                ]

        self.logger.debug(
            '\nFinal consolidated match data set \n{0}\n\n'.format(
                df_match_consol1.shape
                )
            )

        return (df_match_consol1)



"""Define Global variables and constants.

sccmgbls defines all global variables and constants.

See utils for initialization of default values.
"""
import logging

######
# Debug and logging output
######

# Logging levels, default level for logging

LEVELS = {'debug': logging.DEBUG,
          'info': logging.INFO,
          'warning': logging.WARNING,
          'error': logging.ERROR,
          'critical': logging.CRITICAL}

# Logging level

loglvl = logging.INFO

######
# Miscellaneous
######

# Version
VERSION = '1.0'

#   Following is initialized in utils.init_globals()
#   once the working directory is known

#   Working Directories
wkdir = ''
pckdir = ''
confdir = ''
csvdir = ''
nvddir = ''
modeldir = ''

# Scheduling
CANCEL_ON_FAILURE = False
SCHED_TIME = "11:00"
SCHED_SLEEP = 60

######
#   CSV Input data
######

# separate individual values in a particular field
SEP = "|"
SEP2 = ","
HASH = '#'

#######
#   Input data - mostly CSV
#######

v_r_system = ''
v_gs_add_rem_pgms = ''
v_gs_add_rem_pgms_64 = ''
ad_vip_grps = ''
s_vndr_stop_wds = ''
df_label_software = ''
df_label_vendors = ''

######
#   Saved dataframe names
#   'rf_' - refactor code version
######
df_sys_pck = ''
df_add_rem_g_pck = ''
df_cpe4_pck = ''
df_cve_pck = ''
df_v_R_System_3modified_pck = ''
df_sft_vuln_pck = ''
df_match_vendor_publisher_pck = ''
df_match_cpe_sft_pck = ''

######
#   NVD data
######

nvdcpe = ''
nvdcve = ''

# Downloads

url_meta_base = ''
url_meta_end = ''
url_xml_base = ''
url_xml_end = ''
url_cpe = ''

cpe_filename = ''
cpe_max_age = 7

#   Number of years of NVD CVD XML feed files
num_nvd_files = 0

nvd_meta_filename = ''
nvddir = ''

######
#   Models
######
clf_vendor = ''
clf_software = ''

######
#   Fields for the ML classification
######

# Important note: These fields *must* be kept in this order since that is
# how the ML algorithm was originally trained.

#   Vendor data fields
vendor_key_list = ['publisher0', 'vendor_X']
vendor_feature_list = [
                      'fz_ptl_ratio',
                      'fz_ptl_tok_sort_ratio',
                      'fz_ratio',
                      'fz_tok_set_ratio',
                      'fz_uwratio',
                      'ven_len',
                      'pu0_len'
                      ]
vendor_token_list = ['pub0_cln', 'ven_cln']
vendor_attr_list = vendor_key_list + vendor_feature_list + vendor_token_list


#   Software data fields
sft_key_list = [
                'vendor_X',
                'software_X',
                'title_X',
                'DisplayName0',
                'release_X',
                'Version0'
                ]
sft_feature_list = [
                    'fz_ratio',
                    'fz_ptl_ratio',
                    'fz_tok_set_ratio',
                    'fz_ptl_tok_sort_ratio',
                    'fz_uwratio',
                    'fz_rel_ratio',
                    'fz_rel_ptl_ratio',
                    'titlX_len',
                    'DsplyNm0_len'
                    ]

sft_attr_list = sft_key_list + sft_feature_list + ['t_cve_name']

######
#   Plugins
######

#   Plugins must be located in the following directory
PLUGINFOLDER = '/home/jovyan/work/src/plugins'

#   Plugin names specified here must match those in the "*.yapsy-plugin"
#   configuration files

PLUGINIP = 'Plugin Input'
PLUGINRPT = 'Plugin Report'

#   Global plugin manager object
plugin_manager = None



"""matchven: Match NVD CPE "Vendor" data with SCCM "Publisher" data.

Purpose
=======

The matchven module uses heuristics as well as Machine Learning
classification techniques to match NVD CPE "vendor" data to the corresponding
SCCM "Publisher" data originally from the "Add-Remove" entries in the Windows
Registry.

The set of corresponding data is placed in a pandas dataframe which is
persisted.

Public classes
==============

MatchVendor     Match NVD CPE "Vendor" data to SCCM "Publisher" data


                This class uses heuristics and ML Classification to match NIST
                CPE "Vendors" to SCCM "Publishers".

"""

import re
from time import time

import pandas as pd
import numpy as np

import sys

from fuzzywuzzy import fuzz as fz
from sklearn.externals import joblib

import logging

import gbls
import utils
import sccm
import nvd
import ml

# Public classes
__all__ = (
        'MatchVendor'
        )


class MatchVendor(object):
    """Match CPE "Vendor" data to SCCM "Publisher" data.

    The NIST CPE Vendor data is matched to SCCM "Publisher" data.

    *  The input data is normalized, tokenized, and common "stop" words are
       removed.

    *  The Cartesian Product of all possible matches is formed, then reduced
       using simple heuristics.

    *  ML Random Forest classification is used to find matches.

    *  The newly-classified data is concatenated with manually labelled data
       to form the final dataframe of "Vendor" - "Publisher" correspondances.

    Once a reasonably reliable set of Vendor-Publisher corespondances is
    found, this data will be leveraged in the matchsft module to match CPE
    Software titles to SCCM Software inventory data.

    Methods
    -------
    __init__    Class constructor to configure logging, initialize empty data
                frame, and read in "stop" words for the ML Classification.
    load        Load Vendor-Publisher dataframe from the serialized pickled
                file.
    save        Save the Vendor-Publisher dataframe to the corresponding
                serialized, pickled file.
    get         Return a *copy* of the Vendor-Publisher dataframe.
    match       Match the CPE "Vendor" data to the SCCM "Publisher" data

    """

    def __init__(self, mylogger=None):
        """Initialize class by configuring logging,  initializing dataframe.

        This is the class constructor.

        Actions
        -------

        * Initializes logging.
        * Allocates an empty dataframe to contain Vendor-Publisher match data.
        * Reads "stop" words for elimination of common tokens. eg "Inc"

        I/P Parameters
        --------------
        mylogger    logging object. If None, then a new object is initialzed.

        Exceptions
        ----------
        IOError     A message is logged and the error is ignored.

        """
        # Configure logging

        if mylogger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(gbls.loglvl)
        else:
            self.logger = mylogger

        self.logger.info('\n\nInitializing MatchVendor class\n\n')

        self.df_match_vendor_publisher = pd.DataFrame({
            'Match': [],
            'publisher0': [],
            'vendor_X': []
            })

        # "stop" words to improve token matching

        self.__lst_stop_wds = []

        # ML classification

        self.__ML = ml.MLClassify('vendor')

        # Initialize algorithm for ML matching
        # Read stop words

        try:
            s_stop_wds = pd.io.parsers.read_csv(
                                gbls.s_vndr_stop_wds,
                                sep=gbls.SEP2,
                                error_bad_lines=False,
                                warn_bad_lines=True,
                                index_col=0,
                                squeeze=True,
                                header=None,
                                encoding='utf-8')
        except IOError as e:
            self.logger.critical('\n\n***I/O error({0}): {1}\n\n'.format(
                        e.errno, e.strerror))

        except:
            self.logger.critical(
                '\n\n***Unexpected error: {0}\n\n'.format(
                    sys.exc_info()[0]))
            raise

        self.__lst_stop_wds = s_stop_wds.tolist()

        self.logger.info(
                        '\n\nList of stop words: \n{0}\n\n'.format(
                            self.__lst_stop_wds
                            )
                        )

        return None

    def load(self, mypck=None):
        """Load Vendor-Publisher dataframe that was previously saved."""
        self.logger.info(
                '\n\nLoading saved Vendor-Publisher match data into '
                'MatchVendor.df_match_vendor_publisher dataframe\n\n'
                )
        if mypck is None:
            mypck = gbls.df_match_vendor_publisher_pck

        self.df_match_vendor_publisher = pd.read_pickle(mypck)
        return None

    def save(self):
        """Save Vendor-Publisher dataframe in serialized pickle format."""
        self.logger.info(
                    '\n\nSaving MatchVendor.df_match_vendor_publisher '
                    'dataframe\n\n'
                    )
        self.df_match_vendor_publisher.to_pickle(
                        gbls.df_match_vendor_publisher_pck
                        )
        return None

    def get(self):
        """Return a *copy* of the dataframe."""
        df_tmp = self.df_match_vendor_publisher.copy()
        self.logger.info(
                '\n\nGet MatchVendor.df_match_vendor_publisher: '
                '\n{0}\n{1}\n\n'.format(
                                df_tmp.shape,
                                df_tmp.columns
                                )
                )
        return df_tmp

    def match(self, df_cpe, df_sccm_ar):
        """Match the CPE "Vendor" data to the SCCM "Publisher" data.

        Actions
        -------

        *  The input string data is normalized and then converted into
           "tokens" and common "stop" words are removed.

        *  The cartesian product of both sets of data is formed:
           {CPE "Vendor" names} X {SCCM "Publisher" Names}

            - Both I/P dataframes are iterated through to form the set of all
              possible matches.
            - Simple heuristics are used to eliminate improbable matches.
            - "Fuzzy matching" statistics are calculated as ML "features".

        *  Manually labelled match data is read in. The new data to be
           classified is updated using this labelled data. This ensures that
           the ML Classification will run only on new unknown data.

        *  The Random Forest Classification algorithm is run.

        *  The newly-classifed test data is concatenated with the manually
           labelled data to form the final dataframe of Vendor - Publisher
           correspondances.

        Exceptions
        ----------
        IOError     Log error message and ignore

        Returns
        -------
        None

        """
        # Suppress tokens which match "stop" words, e.g. "Inc, Ltd"

        def _remove(x):

            result = ' '
            for str in x:
                if not(str in self.__lst_stop_wds) and (len(str) > 1):
                    result = result + ' ' + str
            return result

        # Prepare the data for ML Classification

        def _match_prepare(df_cpe, df_sccm_ar):

            self.logger.info('\n\nEntering match_prepare\n\n')

            self.logger.info(
                '\n\nMatching NVD vendors to SCCM publisher data\n\n'
                )
            # The I/P CPE 'Vendor' strings along with the SCCM registry
            #   'Add_Remove_Pgms' 'Publisher0' strings are normalized

            # Extract and normalize list of SCCM WMI publishers
            df_arPub = pd.DataFrame(
                            df_sccm_ar['Publisher0'].dropna().str.replace(
                                ' ', '_'
                                ).str.lower().unique(),
                            columns=['publisher0']
                            )

            self.logger.info(
                            '\n\nCount of SCCM Publisher0: {0}\n\n'.format(
                                    df_arPub.shape
                                    )
                            )

            # Extract and normalize list of CPE vendors
            df_cpeVen = pd.DataFrame(
                            df_cpe['vendor_X'].dropna().str.replace(
                                ' ', '_'
                                ).str.lower().unique(),
                            columns=['vendor_X']
                            )
            self.logger.info(
                            '\nCount of CPE vendors: \n{0}\n\n'.format(
                                    df_cpeVen.shape
                                    )
                            )

            # Tokenize names

            # Note:
            #
            # a) CPE data has "underscore" instead of "blank"
            #
            # b) For WMI data, hyphens / ampersands are left "as-is" since
            # these can form part of a vendor name

            df_arPub['pub0_toks'] = df_arPub['publisher0'].str.replace(
                                        '[_.,()+!]', ' '
                                        ).str.split()
            df_cpeVen['vend_toks'] = df_cpeVen['vendor_X'].str.replace(
                                        '[-_]', ' '
                                        ).str.split()

            # Produce new clean names from tokens, taking into account "stop"
            # words

            df_arPub['pub0_cln'] = df_arPub['pub0_toks'].apply(_remove)
            df_cpeVen['ven_cln'] = df_cpeVen['vend_toks'].apply(_remove)

            # show results of tokenization
            self.logger.info(
                '\n\nSample CPE vendors \n{0}\n\n'.format(
                    df_cpeVen[['vendor_X', 'vend_toks', 'ven_cln']].sample(10)
                    )
                )

            self.logger.info(
                '\n\nSample SCCM publishers \n'
                '{0}\n\n'.format(
                    df_arPub[
                        ['publisher0', 'pub0_toks', 'pub0_cln']].sample(10)
                    )
                )

            return (df_arPub, df_cpeVen)

        # Form the Cartesian product of WMI Publishers X CPE vendors

        def _cartesian_product(p_df_arPub, p_df_cpeVen):

            self.logger.info('\n\nEntering cartesian_product\n\n')

            # Force copy-by-value
            df_arPub = p_df_arPub.copy()
            df_cpeVen = p_df_cpeVen.copy()

            # List of name tuples to check
            lst_dict = []
            t0 = time()
            mycount = 0
            self.logger.info(
                            '\n\nStarting generation of '
                            'cartesian product of NIST vendors '
                            'to SCCM publishers ... \n'
                            ' *** This can take some time '
                            '- up to 20 min for large prod datasets.\n\n'
                            )

            # Build the the cartisan product of the two sets of names (CPE,
            # SCCM/WMI) by iterating through the input dataframes

            for (
                    t_cpeix,
                    t_cpeVen_orig,
                    t_cpevend_toks,
                    t_cpeVen
                    ) in df_cpeVen.itertuples():

                #   Ignore cpe vendors that are 1 character long (e.g. 'X')

                if (len(t_cpeVen) < 2):
                    self.logger.debug('cpeVen too short - continuing\n')
                    continue

                for (
                        t_arix,
                        t_arPub0_orig,
                        t_arpub_toks,
                        t_arPub0
                        ) in df_arPub.itertuples():

                    # quick heuristics:
                    #   a) 1st word of cpe Vendor string has to be in the
                    #            tokenized wmi Publisher0 string somewhere
                    #   b) condensed cpe name has to be shorter than the full
                    #           WMI 'Publisher0' name

                    if len(t_cpeVen) > len(t_arPub0):
                        # self.logger.debug('arPub0 too short - continuing'
                        continue

                    # Look for at least one occurence of one word in cpeVen
                    #       somewhere in arPub
                    if fz.partial_token_set_ratio(
                                            t_cpeVen,
                                            t_arPub0,
                                            force_ascii=False
                                            ) < 100:
                        continue

                    # Calculate fuzzy matching statistics as "features" for
                    # the subsequent ML classification

                    lst_dict.append({
                        'publisher0': t_arPub0_orig,
                        'pub0_cln': t_arPub0,
                        'vendor_X': t_cpeVen_orig,
                        'ven_cln': t_cpeVen,
                        'fz_ratio': fz.ratio(
                                t_cpeVen,
                                t_arPub0),
                        'fz_ptl_ratio': fz.partial_ratio(
                                t_cpeVen,
                                t_arPub0),
                        'fz_tok_set_ratio': fz.token_set_ratio(
                                t_cpeVen,
                                t_arPub0,
                                force_ascii=False),
                        'fz_ptl_tok_sort_ratio': fz.partial_token_sort_ratio(
                                    t_cpeVen,
                                    t_arPub0,
                                    force_ascii=False),
                        'fz_uwratio': fz.UWRatio(
                                t_cpeVen,
                                t_arPub0)
                        })
                    mycount = mycount + 1
                    if mycount % 1000 == 0:
                        self.logger.debug(
                                    '# entries produced: {0}\n'.format(
                                                mycount
                                                )
                                     )

                #     # debug code to shorten loop for testing
                #     if mycount > 1000:
                #         break

                # # debug code to speed thru loops
                # if mycount > 1000:
                #     break

            duration = time() - t0
            self.logger.info(
                    '\n\n*** Done in {0} sec\n\n'.format(
                            duration
                            )
                    )
            df_match = pd.DataFrame(lst_dict)

            if df_match.empty:
                self.logger.info(
                    'Resulting cartesian product is empty.\n\n'
                    )

            else:
                self.logger.info(
                            '\n\n Vendor match dataframe: \nCounts {0},\n'
                            'Columns: {1}\n\n'.format(
                                    df_match.shape,
                                    df_match.columns
                                    )
                            )
            return (df_match)

        # Previously manually labelled data is used to update the new data to
        # be classified. This ensures that only new, unknown data is input
        # into the ML Classification algorithm.

        def _update_with_labelled_data(p_df_match):

            self.logger.info('\n\nEntering update_with_labelled_data\n\n')

            df_match = p_df_match.copy()

            # add in length of names as features
            df_match['ven_len'] = df_match['vendor_X'].apply(len)
            df_match['pu0_len'] = df_match['publisher0'].apply(len)

            # Read in the data that was manually labelled in order to
            # originally train the ML model.

            try:
                df_match_lbl = pd.io.parsers.read_csv(
                                        gbls.df_label_vendors,
                                        sep=gbls.SEP2,
                                        error_bad_lines=False,
                                        warn_bad_lines=True,
                                        index_col=0,
                                        encoding='utf-8')
            except IOError as e:
                self.logger.critical('I/O error({0}): {1}'.format(
                    e.errno, e.strerror))

            except:
                self.logger.critical(
                    'Unexpected error: {0}'.format(
                        sys.exc_info()[0]))
                raise

            # Normalize the vendor_X field for this data
            df_match_lbl['vendor_X'] = df_match_lbl['vendor_X'].str.replace(
                                            ' ', '_').str.lower()

            self.logger.info(
                            '\n\n Manually labelled dataframe: counts: {0}\n'
                            ', columns: {1}\n'.format(
                                df_match_lbl.shape,
                                df_match_lbl.columns,
                                df_match_lbl['match'].value_counts()
                                )
                            )

            # Update the data to be matched with the results from manual
            # labelling effort. After doing all this hard manual
            # classification effort, why not use the data?

            df_match_upd = self.__ML.upd_using_labelled_data(
                                        df_match,
                                        df_match_lbl
                                        )

            return (df_match_upd, df_match_lbl)

        # Concatentate the manually labelled data with the newly classified
        # data. Drop extraneous columns to form the final dataframe of
        # "Vendor"-"Publisher" corresspondances.

        def _post_process_matched_data(
                    p_df_match_test2,
                    p_df_match_lbl
                    ):

            # concatenate the two sets of classified data: the manual set, and
            # the machine-classified one

            self.logger.info('\n\nEntering post_process_matched_data\n\n')

            # Force call-by-value
            df_match_test2 = p_df_match_test2.copy()
            df_match_lbl = p_df_match_lbl.copy()

            df_match_consol1 = self.__ML.post_process_matched_data(
                                        df_match_test2,
                                        df_match_lbl
                                        )

            # Drop unneeded columns.
            self.df_match_vendor_publisher = df_match_consol1[[
                                                        'publisher0',
                                                        'vendor_X'
                                                        ]]

            self.logger.debug(
                '\n\nMatched vendors-Publishers dataframe \n{0}\n\n'.format(
                    self.df_match_vendor_publisher .shape
                    )
                )

            self.logger.debug(
                            '\n\nSample matches: \n{0}\n\n'.format(
                                    self.df_match_vendor_publisher.sample(20)
                                    )
                            )

            return None

        ######
        # Mainline code for the "match" method
        ######

        # Prepare i/p data frames for matching
        (df_arPub, df_cpeVen) = _match_prepare(df_cpe, df_sccm_ar)

        # Form Cartesian product for potential matches
        df_match = _cartesian_product(df_arPub, df_cpeVen)

        # Update with labelled data
        (df_match_upd, df_match_lbl) = _update_with_labelled_data(df_match)

        # Do the actual machine learning classification
        (df_match_test2, df_match_labelled) = self.__ML.ml_classify(
                                                            df_match_upd
                                                            )

        # Post process matched vendor-Publisher data
        _post_process_matched_data(
                    df_match_test2,
                    df_match_lbl
                    )

        return None



import sys
import pandas as pd
import re
from yapsy.IPlugin import IPlugin

# modify search path to include parent directory
sys.path.append("../")

import vulns
import gbls

class PluginTwo(IPlugin):
    def print_name(self):
        print (
            'Plugin 2 for custom Vulnmine Reporting'
            ' has loaded successfully.'
            )
        return None

    def custom_stats(self, my_match_vulns):
        """
        Actions
        =======

        Data is grouped by host function to see which type of machine is
        most vulnerable.

        The machines in the VIP AD group are analyzed to determine how
        vulnerable they are.

        Then data is filtered to focus on servers, and then grouped by
        "Region" / criticality of vulns.

        """
        print ('Plugin 2: starting custom_stats.')

        # access the match_vulns dframe
        my_df_sft_vuln = my_match_vulns.get()

        # Have to convert category to original dtype to avoid error
        # in pandas when aggregating / resetting indices.

        my_df_sft_vuln['crit_X'] = my_df_sft_vuln['crit_X_cat'].astype(object)

        # What machines are the most vulnerable?

        # Group data by host function to see which hosts have the most
        # vulnerable software

        dfs2 = my_df_sft_vuln.groupby(['HostFn_X', 'crit_X']).size()
        dfs2a = dfs2.unstack('crit_X').fillna(0)

        print(
            '\n\nGrouping host vuln data by host function \n{0}\n\n'.format(
                                                                    dfs2a
                                                                    )
            )

        dfs2a.reset_index().to_json(
                    orient="records",
                    force_ascii=False,
                    path_or_buf="{0}".format(
                                        gbls.csvdir +
                                        "host_vuln_by_fn.json"
                                        )
                    )

        ######
        # How vulnerable are the machines in the VIP AD group?
        ######

        # How many hosts are in the AD group(s)?
        s_tmp = my_df_sft_vuln['VIP_X'].value_counts()

        print(
            '\n\nCount of hosts in various AD groups\n{0}\n\n'.format(
                        s_tmp
                        )
            )

        s_tmp.reset_index().to_json(
                orient="records",
                force_ascii=False,
                path_or_buf="{0}".format(
                                    gbls.csvdir +
                                    "num_hosts_AD_gps.json"
                                    )

            )

        # Consider only hosts in AD groups
        # Look at vulnerable software for these hosts
        try:
            df_sys_dsA1_vip = my_df_sft_vuln[
                                        my_df_sft_vuln['VIP_X'].notnull()
                                        ]

            # Group by AD grp / Vuln criticality
            df_3sft_most_vuln_excl_gp = df_sys_dsA1_vip.groupby([
                                                            'VIP_X',
                                                            'crit_X_cat'
                                                            ])
            s_tmp = df_3sft_most_vuln_excl_gp.get_group((
                                    'vip',
                                    'High'
                                    ))[
                                        't_cve_name'
                                        ].value_counts().nlargest(25)

            print(
                '\n\nFor hosts in VIP AD group: '
                'what is the most vulnerable software? \n{0}\n\n'.format(
                                        s_tmp
                                        )
                )
            s_tmp.reset_index().to_json(
                orient="records",
                force_ascii=False,
                path_or_buf="{0}".format(
                                    gbls.csvdir +
                                    "VIP_high.json"
                                    )
                )

        except KeyError as e:
            print(
                '\n\n***VIP / High - KeyError: {0}\n\n'.format(e)
                )

        # What about server software? What's vulnerable on servers?
        # Filter data for servers and group by region / vuln criticality

        # Consider only servers
        df_sys_dsA1_tmp = my_df_sft_vuln[
                                my_df_sft_vuln.HostFn_X.notnull()
                                ]
        df_sys_dsA1_srv = df_sys_dsA1_tmp[
                                df_sys_dsA1_tmp.HostFn_X == 'server'
                                ]

        # Group servers by zone
        df_3sft_most_vuln_srv_gp = df_sys_dsA1_srv.groupby([
                                                        'Region_X',
                                                        'crit_X_cat'
                                                        ])

        # List top vulnerable software on 'REGION_A' servers

        try:
            s_tmp = df_3sft_most_vuln_srv_gp.get_group((
                                        'Region_A',
                                        'High'))[
                                            't_cve_name'
                                            ].value_counts().nlargest(25)

            print(
                '\n\nFor servers: list top vulnerable '
                'software for Region_A \n{0}\n\n'.format(s_tmp)
                )

        except KeyError as e:
            print(
                '\n\n***Corp servers - High vulns'
                ' - KeyError: {0}\n\n'.format(e)
                )

        return None



"""Process, parse, persist NVD vendor and software data.

Purpose
=======

The nvd module ingests the following NIST NVD data:
    * CPE dictionary of vendor and software data.
    * CVE vulnerability data including CVSS data.

The data is parsed and placed in pandas dataframes.

The public classes are container objects used to input, persist, parse, and
process NVD data.

Public classes
==============

NvdCpe          NVD CPE vendor / software dictionary data

                The NVD CPE data base lists standardized vendor names /
                software names & versions. These are referenced by the CVE
                data which lists the actual vulnerabilities.


NvdCve          NVD CVE vulnerability data

"""
import re

import pandas as pd
import sys
import datetime
import time
import os

import requests
import xmltodict as xd
import logging

import gbls
import utils

# Public classes
__all__ = (
        'NvdCpe',
        'NvdCve'
        )


class NvdCpe(object):
    """Input, parse, persist NIST NVD vendor/software data.

    The NIST NVD Official Common Platform Enumeration (CPE) Dictionary is a
    structure dataset containing software products published by each vendor.

    From https://nvd.nist.gov/cpe.cfm:

        CPE is a structured naming scheme for information technology systems,
        software, and packages. Based upon the generic syntax for Uniform
        Resource Identifiers (URI), CPE includes a formal name format, a
        method for checking names against a system, and a description format
        for binding text and tests to a name.

    Following is a typical entry from the CPE XML flat file:

    ::

        <cpe-item name="cpe:/a:oracle:jdk:1.7.0:update_60">

            <title xml:lang="en-US">Oracle JDK 1.7.0 Update 60</title>

            <references>
                <reference
                    href="http://www.oracle.com/technetwork/topics/security
                        /cpujul2014-1972956.html">
                    Oracle July 2014 CPU
                </reference>
            </references>

            <cpe-23:cpe23-item
                    name="cpe:2.3:a:oracle:jdk:1.7.0:update_60:*:*:*:*:*:*"/>

        </cpe-item>

    **TL;DR: The CPE dictionary is an XML flat file which gives standardized
    vendor and software names. It also lists the software published by each
    vendor.**

    The I/P data is extracted from the XML flat file and placed in a pandas
    dataframe.

    Methods
    -------
    __init__    Class constructor to configure logging, initialize empty data
                frame

    download_cpe    Download NVD CPE XML dictionary data from the NIST website

    read        Input the NVD CPE dictionary XML file. Clean data and
                remove columns. Extract the nested XML data to form a
                simple pandas dataframe.

    load        Load CPE dataframe from the serialized pickled file.
    save        Save the CPE dataframe to the corresponding pickled file.
    get         Return a *copy* of the CPE dataframe.

    """

    def __init__(self, mylogger=None):
        """Initialize class by configuring logging,  initializing dataframe.

        This is the class constructor. It initializes logging and allocates
        an empty dataframe to contain sccm hosts data.

        I/P Parameters
        --------------
        mylogger    logging object. If None, then a new object is initialzed.

        """
        # Configure logging

        if mylogger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(gbls.loglvl)
        else:
            self.logger = mylogger

        self.logger.info('\n\nInitializing NvdCpe class\n\n')

        self.df_cpe4 = pd.DataFrame({
            #   Fields in NVD software record
            '@name': [],
            'check': [],
            'cpe-23:cpe23-item': [],
            'notes': [],
            'references': [],
            'title': [],
            'cpe23-item-name': [],

            #   Fields added during processing
            'title_X': [],
            'vendor_X': [],
            'software_X': [],
            'release_X': []
            })

    def download_cpe(self):
        """Download NIST CPE XML dictionary data.

        The CPE dictionary lists known vendors / products in a normalized
        format.

        The NVD CVE ("Common Vulnerabilities and Exposures") XML flat files
        use this dictionary to reference vendors / software data.

        Actions
        -------

            Determine current time. Allocate download directory if it does not
            exist.

            if the file currently exists and is too old, then download / unzip
            a new copy.

        Exceptions
        ----------
        RequestException:   The requests module, used for https access, has
                            several exception conditions.

        Returns
        -------
        None

        """

        self.logger.info('\n\nEntering NvdCpe.download_cpe\n\n')

        # Determine current time
        now = time.time()

        do_download = False
        my_cpe = gbls.nvddir + gbls.cpe_filename

        self.logger.debug(
                '\nDownload of cpe file: \n{0}\n\n'.format(
                                            my_cpe
                                            )
                )

        # If CPE dictionary file has already been downloaded

        if os.path.isfile(my_cpe):
            # If CPE file is too old, then download a new copy
            cpe_timestamp = os.path.getmtime(my_cpe)
            cpe_age = (now - cpe_timestamp) / (60*60*24)

            do_download = cpe_age > gbls.cpe_max_age
            self.logger.debug(
                    '\nDownload cpe? \n{0}\n{1}\n\n'.format(
                                                cpe_age,
                                                do_download
                                                )
                    )
        else:
            do_download = True

        if do_download:
            self.logger.info(
                '\nDo CPE download\n\n')

            (cpe_filename, cpe_filecontents) = utils.get_zip(gbls.url_cpe)
            if cpe_filename:
                output_cpe = open(my_cpe, 'w')
                output_cpe.write(cpe_filecontents)
                output_cpe.close()

        return None

    def read(self, my_cpe=None):
        """Read the CPE XML file, parse, and store in pandas dataframe.

        Actions
        -------

        The NVD CPE XML flat file is read. This file documents vendors and
        corresponding published software in a formal, standardized format.

        * The XML file is parsed into a python dictionary. This in turn is
          loaded into a pandas dataframe.

        * The data is cleaned by removing deprecated entries along with
        accompanying columns.

        * The nested "name" data (in the form of an embedded python
          dictionary) is extracted and converted to a pandas dataframe. This
          new dataframe is concatenated to the original dataframe.

        * Entries pertaining to "OS" and "Hardware" are removed.

        * The nested software title information (in the form of an embedded
          python dictionary) is next accessed. Vendor, software name, and
          release data are extracted using pattern matching. The data is added
          to new columns in the pandas dataframe. If the software is released
          in multiple languages, only the en-US version is kept. All of this
          data is added to the dataframe in new columns.

        Exceptions
        ----------
        IOError:    Log an error message and ignore

        Returns
        -------
        None

        """
        self.logger.info('\n\nEntering NvdCpe.read\n\n')

        if my_cpe is None:
                my_cpe = gbls.nvdcpe

        self.logger.debug(
                        'Reading file {0},\nsep: {1}\n\n'.format(
                                my_cpe,
                                gbls.SEP)
                        )

        # read in the uncompressed NVD XML data
        self.logger.info(
            '\n\nReading NvdCpe dictionary\n{0}\n\n'.format(
                    my_cpe
                    )
                )

        with open(my_cpe) as fd:
            dict_cpe = xd.parse(fd.read())

        # convert the python dictionary to a pandas dataframe
        df_cpe = pd.DataFrame.from_dict(
                    dict_cpe['cpe-list']['cpe-item']
                    )

        self.logger.debug(
                    '\n\nRaw input: NIST CPE data: {0}\n\n'.format(
                                df_cpe.shape)
                    )

        # Do an initial cleaning of the data
        # remove deprecated entries
        df_cpe1 = df_cpe[~(df_cpe['@deprecated'] == 'true')]

        self.logger.info(
                    '\n\nCPE vendors left after removing '
                    'depecrated entries: \n{0}\n\n'.format(
                            df_cpe1.shape
                            )
                    )

        # drop the corresponding columns
        df_cpe1.drop(
                ['@deprecated', '@deprecation_date'],
                axis=1,
                inplace=True
                )

        # Extract the embedded 'name' data
        # extract embedded dictionary that has cpe 2.3 name data
        s_cpe_name_dict = df_cpe1['cpe-23:cpe23-item']

        # and convert this to a dataframe
        df_cpe_name = pd.DataFrame(s_cpe_name_dict.tolist())

        # rename column
        df_cpe_name = df_cpe_name.rename(
                columns={'@name': 'cpe23-item-name'}
                )

        # concatenate the two dframes by columns
        df_cpe2 = pd.concat(
            [df_cpe1.reset_index(drop=True), df_cpe_name],
            axis=1,
            join='outer')

        self.logger.debug(
                    '\n\nExtract embedded product name '
                    'data and outer join directly to the '
                    'NIST CPE DataFrame: \n{0} \n {1}\n\n'.format(
                        df_cpe2.shape,
                        df_cpe2.columns
                        )
                    )

        # Remove entries pertaining to 'OS' and 'Hardware'
        # look at applications only. Eliminate 'h' (hardware), 'o' OS
        pattern = re.compile('cpe:2.3:[oh]:', re.IGNORECASE | re.UNICODE)

        df_cpe3 = df_cpe2[
                ~(df_cpe2['cpe23-item-name'].str.contains(pattern))
                ]

        self.logger.info(
                        '\n\nUpdated vendor dataframe after '
                        'removing OS & Hardware entries: \n{0}\n\n'.format(
                            df_cpe3.shape
                            )
                        )

        # Extract embedded software title text
        # extract title text (which is also an embedded dictionary in an ... )
        s_cpe_title_dict = df_cpe3['title']

        def myfn4(row):

            # handle case of software with name in multiple languages

            if isinstance(row, list):
                for elt in row:
                    if elt['@xml:lang'] == 'en-US':
                        return elt['#text']
            else:
                return row['#text']

        df_cpe3['title_X'] = s_cpe_title_dict.apply(myfn4)

        # Extract vendor, software, release information

        pattern = re.compile(
            'cpe:2.3:a:'
            '(?P<vendor_X>[^:]*):'
            '(?P<software_X>[^:]*):'
            '?(?P<release_X>[^:]*):',
            re.IGNORECASE | re.UNICODE)

        df_tmp = df_cpe3['cpe23-item-name'].str.extract(
                                    pattern,
                                    expand=False)

        # add the new columns to the main dataframe
        self.df_cpe4 = pd.concat([df_cpe3, df_tmp], axis=1, join='outer')

        self.logger.debug(
                        '\n\nExtract vendor, software, release data '
                        'and add new columns to the vendor '
                        'dataframe: \n{0} \n{1}\n\n'.format(
                            self.df_cpe4.shape,
                            self.df_cpe4.columns
                            )
                        )

        self.logger.info(
                        '\n\nMajor vendor information: \n{0}\n\n'.format(
                                self.df_cpe4.vendor_X.value_counts().head(10)
                                )
                        )

        self.logger.info(
                        '\n\nMajor software information: \n{0}\n\n'.format(
                            self.df_cpe4.software_X.value_counts().head(10)
                            )
                        )

        return None

    def load(self, mypck=None):
        """Load CPE dataframe that was previously saved."""
        self.logger.info(
                '\n\nLoading saved cpe data into '
                'NvdCpe.df_cpe4 dataframe\n\n')

        if mypck is None:
            mypck = gbls.df_cpe4_pck

        self.df_cpe4 = pd.read_pickle(mypck)
        return None

    def save(self):
        """Save CPE dataframe in serialized pickle format."""
        self.logger.info('\n\nSaving NvdCpe.df_cpe4 dataframe\n\n')
        self.df_cpe4.to_pickle(gbls.df_cpe4_pck)
        return None

    def get(self):
        """Return a *copy* of the dataframe."""
        df_tmp = self.df_cpe4.copy()
        self.logger.info(
                '\n\nGet NvdCpe.df_cpe4: \n{0}\n{1}\n\n'.format(
                                df_tmp.shape,
                                df_tmp.columns
                                )
                )
        return df_tmp


class NvdCve(object):
    """Input, parse, persist NIST CVE vulnerability data.

    The NIST National Vulnerability Database ("NVD") is:

        "the U.S. government repository of standards based vulnerability
        management data."

    From https://nvd.nist.gov/download.cfm:

        XML Vulnerability Feeds - security related software flaws contained
        within XML documents. Each vulnerability in the file includes a
        description and associated reference links from the CVE dictionary
        feed, as well as a CVSS base score, vulnerable product configuration,
        and weakness categorization.

    "CVE" == "Common Vulnerabilities and Exposures". Each CVE entry describes
    "a known vulnerability. Included in the CVE entry are the CVSS scores.

    "CVSS" == "Common Vulnerability Scoring System". This is a set of metrics
    "to assess the severity / impact of a security vulnerability.

    Following is a typical NVD CVE entry from the XML flat file:

    ::

        <entry id="CVE-2015-1683">

            <vuln:vulnerable-configuration id="http://www.nist.gov/">

                <cpe-lang:logical-test operator="OR" negate="false">
                  <cpe-lang:fact-ref name="cpe:/a:microsoft:office:2007:sp3"/>
                </cpe-lang:logical-test>

            </vuln:vulnerable-configuration>

            <vuln:vulnerable-software-list>

                <vuln:product>
                    cpe:/a:microsoft:office:2007:sp3
                </vuln:product>

            </vuln:vulnerable-software-list>

            <vuln:cve-id>CVE-2015-1683</vuln:cve-id>

            <vuln:published-datetime>
                2015-05-13T06:59:14.880-04:00
            </vuln:published-datetime>

            <vuln:last-modified-datetime>
                2015-05-13T11:57:28.013-04:00
            </vuln:last-modified-datetime>

            <vuln:cvss>
              <cvss:base_metrics>

                <cvss:score>9.3</cvss:score>
                <cvss:access-vector>NETWORK</cvss:access-vector>
                <cvss:access-complexity>MEDIUM</cvss:access-complexity>
                <cvss:authentication>NONE</cvss:authentication>

                <cvss:confidentiality-impact>
                    COMPLETE
                </cvss:confidentiality-impact>

                <cvss:integrity-impact>COMPLETE</cvss:integrity-impact>
                <cvss:availability-impact>COMPLETE</cvss:availability-impact>
                <cvss:source>http://nvd.nist.gov</cvss:source>

                <cvss:generated-on-datetime>
                    2015-05-13T11:55:11.580-04:00
                </cvss:generated-on-datetime>

              </cvss:base_metrics>
            </vuln:cvss>

            <vuln:cwe id="CWE-119"/>
            <vuln:references xml:lang="en" reference_type="VENDOR_ADVISORY">
              <vuln:source>MS</vuln:source>

              <vuln:reference
                href="http://technet.microsoft.com/security/bulletin/MS15-046"
                xml:lang="en">

                MS15-046

              </vuln:reference>

            </vuln:references>

            <vuln:summary>
              Microsoft Office 2007 SP3 allows remote attackers to
              execute arbitrary code via a crafted document, aka "Microsoft
              Office Memory Corruption Vulnerability."
            </vuln:summary>

        </entry>

    **TL;DR: The NIST NVD is an standards-based repository of vulnerabilities.
    The vendor and software names can be found in the NIST CPE dictionary.**

    The I/P XML data is parsed, relevant data is extracted and then placed in
    a pandas dataframe.

    Methods
    -------
    __init__    Class constructor to configure logging, initialize empty data
                frame

    download_cve    Download NVD CVE XML feed data from the NIST website

    read        Input the NVD CVE data from the raw XML file. Clean data and
                remove columns. Extract the nested XML data to form a
                simple pandas dataframe.

    load        Load CVE dataframe from the serialized pickled file.
    save        Save the CVE dataframe to the corresponding pickled file.
    get         Return a *copy* of the CVE dataframe.

    """

    def __init__(self, mylogger=None):
        """Initialize class by configuring logging,  initializing dataframe.

        This is the class constructor. It initializes logging and allocates
        an empty dataframe to contain sccm hosts data.

        I/P Parameters
        --------------
        mylogger    logging object. If None, then a new object is initialzed.

        """
        # Configure logging

        if mylogger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(gbls.loglvl)
        else:
            self.logger = mylogger

        self.logger.info('Initializing NvdCve class')

        self.df_cve = pd.DataFrame({
            #   Fields in NVD software record
            'vuln:cve-id': [],
            'vuln:product': [],
            'cvss:access-complexity': [],
            'cvss:access-vector': [],
            'cvss:authentication': [],
            'cvss:availability-impact': [],
            'cvss:confidentiality-impact': [],
            'cvss:integrity-impact': [],
            'cvss:score': [],
            # 170118 Bug fix - Key error, sometimes not present
            #            'vuln:security-protection': [],
            'cvss:source': []

            #   Fields added during processing
            })

    def download_cve(self):
        """Download NIST CVE XML feed data and store in local directory.

        The NVD CVE ("Common Vulnerabilities and Exposures") XML flat files
        are downloaded. These files list known vulnerabilities.

        The data is supplied as a series of files - one file for each
        year. As vulnerability information is updated, files from previous
        years can be updated depending on when the vulnerability was
        discovered.

        For each year, there is also a small file with "meta" data
        describing the main XML file: Time of last update, file size, hash
        of file contents. NIST's intention is apparently twofold: a) limit
        B/W requirements by avoiding downloads of files that have not
        changed, b) protect integrity of downloaded data.

        Actions
        -------

            Determine current year. Allocate download directory if it does not
            exist.

            For each year to be processed:

                Download that year's meta file.

                Compare meta file contents with previous meta file (if it
                exists)

                If meta file contents have changed, then download the updated
                XML Feed file.

                The XML file is unzipped and stored in the download directory.

        Exceptions
        ----------
        RequestException:   The requests module, used for https access, has
                            several exception conditions.

        Returns
        -------
        None

        """

        # Determine current year
        now = datetime.datetime.now()
        my_yr = now.year

        # Process cve files for last "n" years

        for index in range(gbls.num_nvd_files):
            yr_processed = my_yr - index
            self.logger.info(
                '\n\nProcessing NVD files for {0}\n'.format(yr_processed)
                )

            # get the meta file for the year being processed
            url_meta = (
                        gbls.url_meta_base
                        + str(yr_processed)
                        + gbls.url_meta_end
                        )
            self.logger.info(
                '\nReading meta file: \n{0}\n\n'.format(
                                url_meta
                                )
                )
            try:
                resp = requests.get(url_meta)

            except requests.exceptions.RequestException as e:
                    self.logger.critical(
                        '\n\n***NVD XML feeds - Error: '
                        '\n{0}\n{0}\n\n'.format(
                            url_meta,
                            e
                            )
                        )
                    continue

            meta_filename = (
                            gbls.nvddir
                            + gbls.nvd_meta_filename
                            + str(yr_processed)
                            )

            # if file already exists then read the contents
            if os.path.isfile(meta_filename):
                meta_filecontents = open(meta_filename, 'r').read()

                # read updated xml feed file since corresponding meta file
                # contents have changed.

                if meta_filecontents == resp.text:

                    self.logger.info(
                        '\nMeta file unchanged, continuing.\n{0}\n\n'.format(
                                                    meta_filename
                                                    )
                        )
                    continue
                else:
                    self.logger.debug(
                        '\nMeta files differ:\n'
                        '   Current file: {0}\n'
                        '   File read from NVD: {1}\n\n'.format(
                                                    meta_filecontents,
                                                    resp.text
                                                    )
                        )
            else:
                self.logger.debug('\nMeta file does not exist:{0}'.format(
                                                    meta_filename
                                                    )
                                )

            # save new / updated meta file to disk

            output_meta = open(meta_filename, 'w')
            output_meta.write(resp.text)
            output_meta.close()

            # Read the new XML feed file

            url_xml = (
                        gbls.url_xml_base
                        + str(yr_processed)
                        + gbls.url_xml_end
                        )

            (xml_filename, xml_filecontents) = utils.get_zip(url_xml)

            # write this new / updated xml feed file to disk as well

            if xml_filename:

                # hardcode the filenames to avoid problems if NIST changes
                # names

                my_cve_filename = (
                            gbls.nvdcve
                            + str(yr_processed)
                            + '.xml'
                            )

                self.logger.info(
                    '\nSaving XML file I/P {0} as {1}\n\n'.format(
                                                xml_filename,
                                                my_cve_filename
                                                )
                    )

                output_xml = open(my_cve_filename, 'wb')
                output_xml.write(xml_filecontents)
                output_xml.close()

        return None

    def read(self, my_dir=None):
        """Read the CVE XML file, parse, and store in pandas dataframe.

        Actions
        -------

        The NVD CVE ("Common Vulnerabilities and Exposures") XML flat file is
        read. This file contains a list of known vulnerabilities. The venodor
        and software names can be found in the NIST CPE dictionary.

        * The data is supplied in a series of files. There is one file for
          each year. As vulnerability information is updated, files from
          previous years can be updated depending on when the vulnerability
          was discovered.

          Each files is read, parsed into a python dictionary, then converted
          to a pandas dataframe.

          All of these individual data frames are appended to form one
          dataframe.

        * The data is cleaned by eliminating null entries.

        * The nested XML data is painstakingly extracted. The data from the
          corresponding python dictionaries and lists is used to populate new
          columns in the main pandas dataframe.   loaded into a pandas
          dataframe.

        *  The data is cleaned further by removing "OS" and "Hardware"
           entries. Extraneous columns are also dropped.

        *  The CVSS ("Common Vulnerability Scoring System" data in the CVE
           entry is extracted and used to populate additional columns in the
           dataframe.

        Exceptions
        ----------
        IOError:    Log an error message and ignore

        Returns
        -------
        None

        """
        self.logger.info('\n\nEntering NvdCve.read\n\n')

        # Read in the uncompressed NVD XML data
        try:
            fst_time = True

            if my_dir is None:
                my_dir = gbls.nvddir

            # List directory contents

            f = []

            for (dirpath, dirnames, filenames) in os.walk(my_dir):
                f.extend(filenames)
                break

            # Iterate through the cve files

            for my_file in filenames:

                # skip the cpe dictionary if it is there

                if not my_file.startswith(gbls.cve_filename):
                    continue

                my_file1 = my_dir + my_file

                self.logger.info(
                        '\nReading {0}\n\n'.format(
                            my_file1
                            )
                        )

                with open(my_file1) as fd:
                    my_dict = xd.parse(fd.read())

                df_tmp = pd.DataFrame.from_dict(
                            my_dict['nvd']['entry']
                            )
                if fst_time:
                    df_nvd = df_tmp
                    fst_time = False
                else:
                    df_nvd = df_nvd.append(df_tmp)

        except IOError as e:
            self.logger.critical('\n\n***I/O error({0}): {1}\n\n'.format(
                        e.errno, e.strerror))
        except:
            self.logger.critical(
                '\n\n***Unexpected error: {0}\n\n'.format(
                    sys.exc_info()[0]))
            raise

        self.logger.info(
            '\n\nNVD CVE raw data input counts: \n{0}\n{1}\n\n'.format(
                    df_nvd.shape,
                    df_nvd.columns
                    )
            )

        # eliminate null entries. Then reset index to sequential #'s
        df_nvd1 = df_nvd[
                df_nvd['vuln:cvss'].notnull()
                ].reset_index(drop=True)

        self.logger.debug(
            '\n\nNVD CVE data after eliminating '
            'null entries: \n{0}\n\n'.format(
                    df_nvd1.shape
                    )
            )

        # pull out cvss_dict hierarchical entry in each row
        s_cvss_dict = df_nvd1['vuln:cvss'].apply(
                        lambda mydict: mydict['cvss:base_metrics']
                        )

        # convert this series to a dataframe
        df_cvss = pd.DataFrame(s_cvss_dict.tolist())

        # Finally, concatenate the two dataframes into one (by columns)
        df_nvd2 = pd.concat([df_nvd1, df_cvss], axis=1, join='outer')

        self.logger.debug(
            '\n\nNVD CVE counts, columns after '
            'pulling out embedded CVSS data: \n{0}\n{1}\n\n'.format(
                        df_nvd2.shape,
                        df_nvd2.columns
                        )
            )

        # drop unneeded columns
        df_nvd2.drop(
                ['@id', 'vuln:cwe'],
                axis=1,
                inplace=True
                )

        # Build a table of vulns vs software

        # extract list of impacted software for each vulnerability
        # build a new dataframe with the vuln's cve-id
        df_sft = df_nvd2[['vuln:cve-id']]

        # then pull out the embedded list of vulnerable software
        # and if the software list is empty, then handle this gracefully
        df_sft['sftlist'] = df_nvd2[
            'vuln:vulnerable-software-list'].fillna(value='xx').apply(
                lambda mydict: [] if mydict == 'xx' else mydict[
                                                            'vuln:product'
                                                            ]
                )

        self.logger.info(
                '\n\n embedded software counts: \n{0}'.format(
                                df_sft.shape
                                )
                )

        # expand each embedded software list into a list containing tuples of
        # the form ['cve_id', 'software_id']

        # initialize an empty list
        lst_sft = []

        # This fn pulls out the embedded list and builds the tuples
        def myfn3(row):
            mylist = row['sftlist']

            # handle case where there is only 1 vuln software in the list
            if not(isinstance(mylist, list)):
                    mylist = [mylist]

            # append each new tuple onto the end of the existing list
            for sft in mylist:
                lst_sft.append([row['vuln:cve-id'], sft])

            # return some value to keep pandas happy
            return 1

        # Run through the dataframe and apply the fn to each row in turn
        df_sft.apply(myfn3, axis=1)

        # Now convert the list of tuples to a dataframe
        df_sft1 = pd.DataFrame(
                lst_sft,
                columns=['vuln:cve-id', 'vuln:product']
                )

        # remove os/hdware entries. Best to analyze MS OS vulns by using MS'
        # patch csv file

        pattern = re.compile(
                        'cpe:'
                        '/[oh]:',
                        re.IGNORECASE | re.UNICODE
                        )

        df_sft2 = df_sft1[
                        ~df_sft1['vuln:product'].str.contains(pattern)
                        ]

        # Finally add information describing the vulnerability add in cvss
        # information concerning the vuln characteristics and severity

        # pull out the cvss information for each vulnerability
        df_cvss = df_nvd2[[
                'vuln:cve-id',
                u'cvss:access-complexity',
                u'cvss:access-vector',
                u'cvss:authentication',
                u'cvss:availability-impact',
                u'cvss:confidentiality-impact',
                u'cvss:integrity-impact',
                u'cvss:score',
                # 170118 Bug fix: Sometimes not present
                #                u'vuln:security-protection',
                u'cvss:source'
                ]]

        # Now merge it into the new dataframe mapping software to vulns
        self.df_cve = pd.merge(
                        df_sft2,
                        df_cvss,
                        how='inner',
                        on='vuln:cve-id'
                        )

        self.logger.info(
                        '\n\nUpdated software dataframe '
                        'which maps software to vulns: \n{0}\n{1}\n\n'.format(
                                self.df_cve.shape,
                                self.df_cve.columns
                                )
                        )

        return None

    def load(self, mypck=None):
        """Load NvdCve vulnerability dataframe that was previously saved."""
        self.logger.info(
                '\n\nLoading saved CVE data into '
                'NvdCve.df_cve dataframe\n\n'
                )
        if mypck is None:
            mypck = gbls.df_cve_pck

        self.df_cve = pd.read_pickle(mypck)
        return None

    def save(self):
        """Save NvdCpe vuln dataframe in serialized pickle format."""
        self.logger.info('\n\nSaving NvdCve.df_cve dataframe\n\n')
        self.df_cve.to_pickle(gbls.df_cve_pck)
        return None

    def get(self):
        """Return a *copy* of the data."""
        df_tmp = self.df_cve.copy()
        self.logger.info(
                '\n\nGet NvdCve.df_cve: \n{0}\n{1}\n\n'.format(
                                df_tmp.shape,
                                df_tmp.columns
                                )
                )
        return df_tmp



"""vulns: Match NVD vuln data with SCCM host data, output basic stats.

Purpose
=======

The vulns module matches the NIST NVD CVE vulnerability data with the SCCM
host data.

The input CVE vulnerability data is categorized as "Low/Med/High" using a
simplistic mapping of the CVSS metrics.

Then the CVE vuln data is merged with the SCCM Software inventory data. This
is turn is merged with the SCCM Host data to produce a dataframe listing
vulnerable software for each host.

The data is grouped in various ways to produce summary statistics.


Public classes
==============

MatchVulns      Merge CVE Vuln data with SCCM host data, print basic stats.

"""
import re

import pandas as pd

import sys


import logging

import gbls
import utils
import sccm
import nvd

# Public classes
__all__ = (
        'MatchVulns'
        )


class MatchVulns(object):
    """Match NVD CPE "Software" data to SCCM "Software" inventory data.

    Actions
    -------
    The input CVE vulnerability data is categorized as "Low/Med/High" using a
    simplistic mapping of the CVSS metrics.

    Then the CVE vuln data is merged with the SCCM Software inventory data.
    This is turn is merged with the SCCM Host data to produce a dataframe
    listing vulnerable software for each host.

    The data is grouped in various ways to produce summary statistics.

    Methods
    -------
    __init__    The class constructor which initializes logging, and an empty
                dataframe for the resulting matches.

    load        Load merged software vulns dataframe that was
                previously saved in pickled format.

    save        Save merged software vulns dataframe to a
                serialized pickled flat file.

    get         Return a *copy* of the merged software vulns dataframe.

    data_merge  Categorize the I/P CVSS data by using a simplistic mapping of
                the CVSS metrics.

                Merge the SCCM Software inventory data is merged with the CPE
                Software - SCCM Software correspondance data.

                The CVE Vulnerability data is next merged with the above.

                Finally merge the SCCM host data to produce a dataframe that
                lists vulnerable software for each host.

    basic_stats Group / sort the merged software vuln data in various ways to
                produce various summary statistics.
    """

    def __init__(self, mylogger=None):
        """Initialize class by configuring logging,  initializing dataframe.

        This is the class constructor. It initializes logging and allocates
        an empty dataframe to contain sccm hosts data.

        I/P Parameters
        --------------
        mylogger    logging object. If None, then a new object is initialzed.

        """
        # Configure logging

        if mylogger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(gbls.loglvl)
        else:
            self.logger = mylogger

        self.logger.info('\n\nInitializing MatchVulns class\n\n')

        self.df_sft_vuln = pd.DataFrame({
                                'ResourceID': []
                                })

    def load(self, mypck=None):
        """Load merged software vulns data that was previously saved."""
        self.logger.info(
                '\n\nLoading software vulns data into '
                'df_sft_vuln dataframe\n\n'
                )
        if mypck is None:
            mypck = gbls.df_sft_vuln_pck
        self.df_sft_vuln = pd.read_pickle(mypck)
        return None

    def save(self):
        """Save the merged software vulns data in pickle serialized form."""
        self.logger.info('\n\nSaving df_sft_vuln dataframe\n\n')
        self.df_sft_vuln.to_pickle(gbls.df_sft_vuln_pck)
        return None

    def get(self):
        """Return a *copy* of the data."""
        df_tmp = self.df_sft_vuln.copy()
        self.logger.info(
                '\n\nGet MatchVulns.df_sft_vuln:\n{0}\n{1}\n\n'.format(
                                df_tmp.shape,
                                df_tmp.columns
                                )
                )
        return df_tmp

    def data_merge(
                self,
                df_cve,
                df_match_cpe_sft,
                df_add_rem_g,
                df_sys
                ):
        """Match the CPE "Software" data to the SCCM "Software" data.

        Actions
        -------
        *   The input CVE vulnerability data is categorized.

        *   Calculate the worst case for each of the 3 CVSS metrics: score /
            access complexity / access vector.

        *   A simplistic fn uses the CVSS scores to classify the Vuln as "Low
            / Med / High":

                High:   Overall CVSS score > 7
                        Complexity: LOW / MED
                        Access: NETWORD / ADJACENT NETWORK

                Low:    Overall CVSS score < 4 *OR*
                        (Complexity: HIGH *AND* Local Access only)

                Medium: Everything else

        *   Rename some columns for easier access, drop extraneous columns as
            well.

        *   The SCCM Software inventory data is merged with the CPE Software -
            SCCM Software correspondance data.

        *   The CVE Vulnerability data (which is keyed on CPE Software) is
            next merged with the above.

        *   Finally the SCCM host data is merged with the above. This gives a
            dataframe that lists vulnerable software for each host.


        Returns
        -------
        None

        """
        self.logger.info('\n\nEntering match_vulns_sft\n\n')

        # Categorize the CVE vulnerabilities using a simplistic mapping of
        # the corresponding CVSS metrics.

        def _categorize_cvss_data(df_sft3):
            # The CVE vulnerability data is categorized.

            # Use the CVE vulnerability data as input.
            # Categorize the CVSS impact data
            # Calculate maximum impact for each software
            # Convert the "worst case" impact into a simple classification
            # "Hi-Med-Low"

            self.logger.info(
                    '\n\nEntering categorize_cvss_data\n\n')

            # Function to compute simplistic criticality value

            def myfn6(row):
                # self.logger.debug('input: ', row
                try:
                    my_score = row['cvss_score']
                    my_ease = row['cvss_acc_cmpl_cat'] in [
                                                        'LOW',
                                                        'MEDIUM'
                                                        ]
                    my_access = row['cvss_acc_vect_cat'] in [
                                                        'NETWORK',
                                                        'ADJACENT_NETWORK'
                                                        ]
                    if (my_score > 7) and my_ease and my_access:
                        return ('High')

                    elif (my_score < 4) or (
                                    (not my_ease)
                                    and (not my_access)
                                    ):
                        return('Low')

                    else:
                        return('Medium')
                except:
                    return("None")

            # Categorize the CVSS impact data

            # categories
            df_sft3['cvss_acc_cmpl_cat'] = df_sft3[
                        'cvss:access-complexity'].astype(
                            'category',
                            categories=['HIGH', 'MEDIUM', 'LOW'],
                            ordered=True
                            )

            df_sft3['cvss_acc_vect_cat'] = df_sft3[
                            'cvss:access-vector'].astype(
                                'category',
                                categories=[
                                        'LOCAL',
                                        'ADJACENT_NETWORK',
                                        'NETWORK'
                                        ],
                                ordered=True)

            # convert from string to float for max comparisons
            df_sft3['cvss_score'] = pd.to_numeric(
                                        df_sft3['cvss:score'],
                                        errors='coerce'
                                        )

            # get rid of extraneous data columns
            df_sft3.drop(
                    [
                        u'cvss:score',
                        u'cvss:access-complexity',
                        u'cvss:access-vector',
                        u'cvss:authentication',
                        u'cvss:availability-impact',
                        u'cvss:confidentiality-impact',
                        u'cvss:integrity-impact'],
                    # u'vuln:security-protection'],
                    # 170118 Bug fix: Sometimes not present
                    inplace=True,
                    axis=1)

            # rename some other columns for easier access
            df_sft4 = df_sft3.rename(
                                columns={
                                    'vuln:cve-id': 'cve_id',
                                    'vuln:product': 'cpe_prod',
                                    'cvss:source': 'cvss_src'}
                                    )

            self.logger.info(
                    '\n\nProcessing CVE '
                    'vulnerability data: \n{0}\n{1}\n\n'.format(
                            df_sft4.shape,
                            df_sft4.columns
                            )
                    )

            # Calculate maximum impact for each software

            # group vulns by software
            df_sft4_gp = df_sft4.groupby('cpe_prod')

            # compute worst case value for each software
            df_sft4_agg = df_sft4_gp.agg({'cvss_score': max,
                                          'cvss_acc_cmpl_cat': max,
                                          'cvss_acc_vect_cat': max})

            self.logger.debug(
                '\n\n Aggregated CVE vuln data '
                'for worst case\n{0}\n{1} '.format(
                        df_sft4_agg.shape,
                        df_sft4_agg.columns
                        )
                )

            # Convert the "worst case" impact into a simple classification
            # of "Hi-Med-Low"

            # Compute the criticality for the vuln data
            df_sft4_agg['crit_X'] = df_sft4_agg.apply(myfn6, axis=1)

            self.logger.debug(
                    '\nAggregated vuln data counts '
                    'by criticality \n{0}\n\n'.format(
                            df_sft4_agg.crit_X.value_counts()
                            )
                    )

            # The criticality value is converted to a pandas category
            my_crit_categories = ['None', 'Low', 'Medium', 'High']

            # and then convert the calculated value to a category
            df_sft4_agg['crit_X_cat'] = df_sft4_agg['crit_X'].astype(
                                            'category',
                                            categories=my_crit_categories,
                                            ordered=True
                                            )

            df_sft4_agg.drop(['crit_X'], axis=1, inplace=True)

            self.logger.info(
                    '\n\nAggregated vuln data: \n{0}\n{1}\n\n'.format(
                                                        df_sft4_agg.shape,
                                                        df_sft4_agg.columns
                                                        )
                    )

            return (df_sft4_agg)

        # Match vulnerabilities with software inventory data
        # Then merge this with the SCCM host data.

        def _update_hosts_with_vulns(df_sft4_agg):

            self.logger.info(
                    '\n\nEntering update_hosts_with_vulns\n\n')

            ######
            #   Process SCCM software inventory data
            ######

            # drop unused columns
            df_add_rem_g2 = df_add_rem_g.drop(
                                            [
                                                'AgentID',
                                                'GroupID',
                                                'ProdID0',
                                                'RevisionID',
                                                'TimeStamp'
                                                ],
                                            axis=1
                                            )

            self.logger.info(
                '\nSCCM inventory data to '
                'start with: \n{0}\n{1}\n\n'.format(
                        df_add_rem_g2.shape,
                        df_add_rem_g2.columns
                        )
                )

            ######
            # Merge SCCM inventory with "match" data
            ######

            self.logger.info(
                '\nProduct-software match dframe i/p '
                '\n{0}\n{1}\n\n'.format(
                        df_match_cpe_sft.shape,
                        df_match_cpe_sft.columns
                        )
                )

            df_add_rem_g3 = pd.merge(
                                df_add_rem_g2,
                                df_match_cpe_sft,
                                how='inner',
                                on=['DisplayName0', 'Version0']
                                )

            self.logger.debug(
                '\n\nSCCM inventory data after '
                'inner join with product-software '
                'match dataframe: \n{0}\n{1}\n\n'.format(
                        df_add_rem_g3.shape,
                        df_add_rem_g3.columns
                        )
                )

            ######
            # Merge vulnerability data with sccm software inventory data
            ######

            # Merge inventory data with aggregated vuln data
            # Index of df_sft4_agg is the CPE ID.

            self.logger.debug(
                '\nVuln dframe i/p '
                '\n{0}\n{1}\n\n'.format(
                        df_sft4_agg.shape,
                        df_sft4_agg.columns
                        )
                )

            df_add_rem_g4 = pd.merge(
                                df_add_rem_g3,
                                df_sft4_agg,
                                how='inner',
                                left_on='t_cve_name',
                                right_index=True
                                )

            self.logger.info(
                '\n\nSCCM inventory data after '
                'inner join with vulnerability data\n{0}\n{1}\n\n'.format(
                        df_add_rem_g4.shape,
                        df_add_rem_g4.columns
                        )
                )

            # Look at distribution of vuln criticality
            self.logger.info(
                '\nDistribution of vulnerabilities '
                'in installed software\n{0}\n\n'.format(
                    df_add_rem_g4['crit_X_cat'].value_counts()
                    )
                )

            ######
            # Now add in SCCM host data
            ######

            self.logger.debug(
                '\nSCCM discovered '
                'host data\n{0}\n{1}\n\n'.format(
                    df_sys.shape,
                    df_sys.columns
                    )
                )

            # Merge host data with inventory / vulnerability data
            self.df_sft_vuln = pd.merge(
                                    df_sys,
                                    df_add_rem_g4,
                                    how='inner',
                                    on='ResourceID'
                                    )

            self.logger.info(
                '\n\nSCCM host data with '
                'merged vulnerability information\n{0}\n{1}\n\n'.format(
                    self.df_sft_vuln.shape,
                    self.df_sft_vuln.columns
                    )
                )

            return None

        ######
        # mainline code for match_vulns_sft method
        ######

        # Do the matching

        if df_match_cpe_sft.empty:
            self.logger.critical(
                '\n\n*** I/P dframe df_match_cpe_sft empty. '
                'No host-vuln matching will be done.\n\n'
                )
            # initialize to an empty dframe
            self.df_sft_vuln = pd.DataFrame({'K1': []}, index=[])
        else:

            df_sft4_agg = _categorize_cvss_data(df_cve)

            if df_sft4_agg.empty:
                    self.logger.critical(
                        '\n\n*** dframe df_sft4_agg empty. '
                        'No host-vuln matching will be done.\n\n'
                        )
                    # initialize to an empty dframe
                    self.df_sft_vuln = pd.DataFrame({'K1': []}, index=[])
            else:

                _update_hosts_with_vulns(df_sft4_agg)

        return None

    def basic_stats(self):
        """Analyze vulnerable 3rd-party software.

        Actions
        -------

        Vuln / inventory data is grouped by vulnerability criticality. This
        shows which vulnerable software is the most widely-installed.

        Next data is sorted by site / criticality to show where vulnerable
        software is installed.

        Returns
        -------
        None
        """
        self.logger.info('\n\n*** Entering basic_stats\n')

        my_df_sft_vuln = self.get()

        if my_df_sft_vuln.empty:
            self.logger.critical(
                '\n\n*** Empty dframe MatchVulns.df_sft_vuln. '
                'No stats can be produced.\n\n')
            return None

        my_df_sft_vuln.to_csv(
                        sep=gbls.SEP,
                        encoding='utf-8',
                        path_or_buf="{0}".format(
                                            gbls.csvdir +
                                            "sft_vuln_raw.csv"
                                            )
                        )

        # Have to convert category to original dtype to avoid error
        # in pandas when aggregating / resetting indices.

        my_df_sft_vuln['crit_X'] = my_df_sft_vuln['crit_X_cat'].astype(object)


        # Group by vuln criticality, then look at each group
        df_3sft_most_vuln_gp = my_df_sft_vuln.groupby(['crit_X_cat'])

        # top 25 most widely deployed software in "High" group
        s_tmp = df_3sft_most_vuln_gp.get_group(
                        'High'
                        )['t_cve_name'].value_counts().nlargest(25)

        self.logger.info(
            '\n\nTop 25 most widely deployed software '
            'with "High" vulns:\n{0}\n\n'.format(s_tmp)
            )

        s_tmp.reset_index().to_json(
                orient="records",
                force_ascii=False,
                path_or_buf="{0}".format(
                                    gbls.csvdir +
                                    "top25_sft_high.json"
                                    )
                )

        ######
        # Where are these software installed?
        ######

        # Where are these software installed?

        # Group data by site / vuln criticality
        dfs1 = my_df_sft_vuln.groupby(['Site_X', 'crit_X']).size()

        try:

            dfs1a = dfs1.unstack('crit_X').fillna(0).nlargest(
                                                            10,
                                                            ['High', 'Medium']
                                                            )
        except KeyError as e:
            self.logger.critical(
                '\n\n***Site/ High-Med vulns - KeyError: {0}\n\n'.format(e)
                )

        # dfs1a.plot(kind='bar', stacked=True)

        self.logger.info(
            '\n\nGrouping host vuln data '
            'by site / vuln criticality:\n{0}\n\n'.format(dfs1a)
            )

        dfs1a.reset_index().to_json(
                    orient="records",
                    force_ascii=False,
                    path_or_buf="{0}".format(
                                        gbls.csvdir +
                                        "site_by_criticality.json"
                                        )
                    )

        return None



"""Mine SCCM data for 3rd-party vulnerabilities.

Usage:

vulnmine.py [-h] [-d] [-l loglevel] [-w work_directory] [--help] [
            --debug] [--log=loglevel] [--workdir=work_directory]

-h --help   Produce this help messages

-l loglevel --log=loglevel

             Specify the level of logging:
                debug, info, warning, error, critical.
             Logging is to standard output.

-w work_directory --workdir=work_directory

            Specify work directory containing data

Function
--------

Input files are assumed to be in various sub-directories of the main work
directory for this program.

    SCCM data from Powershell scripts -- CSV files
    AD data from PS scripts -- CSV files
    NVD / CPE data from NIST -- zip archives
    Pickled pandas dataframes such as 'stop' words for tokenization
    Machine Learning models in scipy 'joblib' dump format

Outputs:
    Pickled pandas dataframes for
        SCCM data:
            Discovered hosts
            Software Inventory data for SCCM WMI Registry 'Add'
        NIST NVD data:
            CVE/CVSS vulnerability data
            CPE vendor-software database data
    JSON file with summarized, consolidated SCCM / NIST data

Examples
--------

    python vulnmine.py mydir
"""
import logging
import argparse
import sys
import os
import schedule
import time
import functools
from yapsy.PluginManager import PluginManager

import utils
import gbls
import sccm
import nvd
import matchven
import matchsft
import vulns

# Don't export any symbols
__all__ = ()


######
#   Custom Exception Handlers
######

class Usage(Exception):
    def __init__(self, msg):
        self.msg = msg


class NameInvalid(Exception):
    """Handle exception if Working directory invalid."""

######
#   Functions to handle specific processing
######


def rd_sccm_hosts():
    """Read in the SCCM hosts data and munge it"""
    hosts = sccm.SccmHosts()
    hosts.read()
    hosts.save()

    # Invoke Input plugin for customized I/P data
    plugin1 = gbls.plugin_manager.getPluginByName(gbls.PLUGINIP)
    plugin1.plugin_object.modify_hosts(hosts)

def rd_sccm_sft():
    """Read SCCM software inventory data"""
    sft = sccm.SccmSoft()
    sft.read()
    sft.save()

def rd_cpe():
    """Read and process NIST NVD vendor / software data"""
    # This data describes (using a well-defined standardized format)
    # the software products produced by each vendor.

    cpe = nvd.NvdCpe()
    cpe.download_cpe()
    cpe.read()
    cpe.save()

def rd_cve():
    """Read and process NIST Vulnerability data"""

    # This data references the NIST CPE data to describe known
    # vulnerabilities for each software product / version.

    cve = nvd.NvdCve()
    cve.download_cve()
    cve.read()
    cve.save()

def match_vendors():
    """Match CPE vendor data to SCCM publisher data"""
    # Initialize inputs
    cpe = nvd.NvdCpe()
    cpe.load()
    sft = sccm.SccmSoft()
    sft.load()

    # Process
    match_vendor = matchven.MatchVendor()

    match_vendor.match(
                cpe.get(),
                sft.get()
                )
    match_vendor.save()

def match_sft():
    """Match CPE software to SCCM Publishers"""
    #   Use the set of vendor - publisher correspondance data
    #   to determine the possible software for each SCCM publisher.
    #
    #   Then match CPE software data to SCCM software inventory data

    # Initialize inputs
    sft = sccm.SccmSoft()
    sft.load()
    match_vendor = matchven.MatchVendor()
    match_vendor.load()
    cpe = nvd.NvdCpe()
    cpe.load()

    # Process
    match_soft = matchsft.MatchSoft()

    match_soft.match(
            match_vendor.get(),
            sft.get(),
            cpe.get()
            )
    match_soft.save()

def upd_hosts_vulns():
    """Determine vulnerable software installed on each SCCM host"""
    #   Update SCCM host data with consolidated vuln data
    #   Produce some basic statistics

    # Initialize inputs
    hosts = sccm.SccmHosts()
    hosts.load()
    cve = nvd.NvdCve()
    cve.load()
    sft = sccm.SccmSoft()
    sft.load()
    match_soft = matchsft.MatchSoft()
    match_soft.load()

    # Process

    match_vulns = vulns.MatchVulns()

    match_vulns.data_merge(
                cve.get(),
                match_soft.get(),
                sft.get(),
                hosts.get()
                )

    match_vulns.save()

def output_stats():
    match_vulns = vulns.MatchVulns()
    match_vulns.load()

    match_vulns.basic_stats()

    # Invoke Report plugin for customized stats
    plugin2 = gbls.plugin_manager.getPluginByName(gbls.PLUGINRPT)
    plugin2.plugin_object.custom_stats(match_vulns)

def do_all():
    rd_sccm_hosts()
    rd_sccm_sft()
    rd_cpe()
    rd_cve()
    match_vendors()
    match_sft()
    upd_hosts_vulns()
    output_stats()


######
#   Mainline code
######

def main():
    """Handle case where pgm is called as a script from the cmd line."""
    parser = argparse.ArgumentParser(
        description='Process SCCM, NVD data.'
        )

    parser.add_argument(
        '-l', '--loglevel',
        choices=[
            'debug',
            'info',
            'warning',
            'error',
            'critical'
            ],
        help='Set desired verbosity for logging.'
        )

    parser.add_argument(
        '-a', '--action',
        default='sched',
        nargs='+',
        choices=[
            'rd_sccm_hosts',
            'rd_sccm_sft',
            'rd_cpe',
            'rd_cve',
            'match_vendors',
            'match_sft',
            'upd_hosts_vulns',
            'output_stats',
            'all'
            ],
        help='Desired action to perform.'
        )

    parser.add_argument(
        '-w', '--workdir',
        default='/home/jovyan/work/',
        help='Specify work directory.'
        )

    parser.add_argument(
        '-y', '--years',
        type=int,
        default=5,
        help="Number of CVE files to download - 1 file for each year's data."
        )

    parser.add_argument(
        '--version',
        action='version',
        version=gbls.VERSION
        )

    args = parser.parse_args()

    # Set working directory

    if not os.path.exists(args.workdir):
        print("{0} does not exist".format(args.workdir))
        raise NameInvalid("Directory does not exist")

    elif not os.path.isdir(args.workdir):
        raise NameInvalid("Element is not a directory")

    else:
        gbls.wkdir = args.workdir

    # Initialize global variables now that the working directory
    # is known
    utils.init_globals()

    # initialize logging
    utils.setup_logging()

    if args.loglevel is not None:
        gbls.loglvl = gbls.LEVELS.get(
            args.loglevel,
            logging.NOTSET
            )

    logging.basicConfig(level=gbls.loglvl)
    logger = logging.getLogger(__name__)
    logger.setLevel(gbls.loglvl)

    logger.debug(
        "\nLog level set to {0}\n\n".format(gbls.loglvl)
        )

    # Set number of years of CVE files to download
    gbls.num_nvd_files = args.years

    logger.info(
        '\n{0} CVE files (1 per yr of data) '
        'will be downloaded.\n\n'.format(gbls.num_nvd_files)
        )

    # Check that I/P directories are available

    if not os.path.isdir(gbls.confdir):
        logger.critical('***"conf" dir is missing. Execution terminated.\n\n')
        return(100)
    if not os.path.isdir(gbls.csvdir):
        logger.critical('***"csv" dir is missing. Execution terminated.\n\n')
        return(100)
    if not os.path.isdir(gbls.modeldir):
        logger.critical(
            '***"model" dir is missing. Execution terminated.\n\n'
            )
        return(100)

    logger.info(
        '\n\nWorking directory: {0}\n'
        'Action to perform: {1}\n'.format(
                                        gbls.wkdir,
                                        args.action
                                        )
        )

    # Load plugins

    utils.load_plugins()

    ######
    #   Scheduling functions
    ######

    def catch_exceptions(job_func):
        @functools.wraps(job_func)
        def wrapper(*args, **kwargs):
            try:
                return job_func(*args, **kwargs)
            except:
                import traceback
                print(traceback.format_exc())
                if gbls.CANCEL_ON_FAILURE:
                    return schedule.CancelJob
        return wrapper

    @catch_exceptions
    def run_job():
        logger.info(
            '\n\n'
            '*** Scheduled run starting. ***\n'
            '*** Scheduled run starting. ***\n'
            '*** Scheduled run starting. ***\n'
            '*** Scheduled run starting. ***\n'
            '\n\n'
            )
        do_all()

    #######
    #    Main Processing Loop
    ######

    # Perform specified actions

    if 'sched' in args.action:
        # Run scheduling loop to execute once daily

        # schedule.every(1).minutes.do(run_job)
        schedule.every().day.at(gbls.SCHED_TIME).do(run_job)

        while True:
            schedule.run_pending()
            time.sleep(gbls.SCHED_SLEEP)

    elif 'all' in args.action:
        do_all()

    else:
        for my_action in args.action:

            if my_action == 'rd_sccm_hosts':
                rd_sccm_hosts()

            elif my_action == 'rd_sccm_sft':
                rd_sccm_sft()

            elif my_action == 'rd_cpe':
                rd_cpe()

            elif my_action == 'rd_cve':
                rd_cve()

            elif my_action == 'match_vendors':
                match_vendors()

            elif my_action == 'match_sft':
                match_sft()

            elif my_action == 'upd_hosts_vulns':
                upd_hosts_vulns()

            elif my_action == 'output_stats':
                output_stats()

            else:
                logger.critical(
                            "***Error vulnmine: action: {0}\n\n".format(
                                                            my_action
                                                            )
                            )

if __name__ == '__main__':
    sys.exit(main())



"""matchsft: Match CPE "Software" data with SCCM Software Inventory data.

Purpose
=======

The matchven module previously produced a dataframe with CPE "Vendor" - SCCM
"Publisher" matches. This dataframe is used by matchsft as follows:

*   The CPE dictionary documents the software produced by each Vendor.
    Likewise, SCCM Sotware Inventory contains the set of software installed
    for each SCCM "Publisher".

*   For a given Vendor-Publisher pair, the Cartesiasn Product of the two
    respective software sets is formed to give the set of all possible
    matching software.

*   This data set is reduced using simple heuristics, refined using manually
    labelled data, and then input to ML Classification.


The result is a dataframe giving matches NVD CPE "Software" data to the
corresponding SCCM "Software" data originally from the "Add-Remove" entries in
the Windows Registry.


Public classes
==============

MatchSoft     Match NVD CPE "Software" data to SCCM "Software" inventory data

              This class uses heuristics and ML Classification to match NVD
              CPE "Software" data to the corresponding SCCM "Software"
              inventory data.

"""
import re
from time import time

import pandas as pd
import numpy as np

import sys

from fuzzywuzzy import fuzz as fz
from sklearn.externals import joblib

import logging

import gbls
import utils
import sccm
import nvd
import ml

# Public classes
__all__ = (
        'MatchSoft'
        )


class MatchSoft(object):
    """Match NVD CPE "Software" data to SCCM "Software" inventory data.

    Actions
    -------

    *  Both the SCCM software inventory data and the CPE Sofware data are
       cleaned, and normalized. Extraneous columns are dropped. Missing
       values are initialized. Both sets of input data are sorted / grouped.

    *  The cartesian product is formed for each Vendor-Publisher tuple:
       {CPE "Software" for this vendor} X
            {SCCM Software installed for this "Publisher"}

       Simple heuristics are used to reduce # potential matches. "Fuzzy
       matching" statistics are calculated as ML "features".

    *  Manually labelled match data is read in. The new data to be
       classified is updated using this labelled data. This ensures that only
       new unclassified data is input to ML classification.

    *  The Random Forest Classification algorithm is run.

    *  The newly-classifed test data is concatenated with the manually
       labelled data to form the final dataframe of CPE-SCCM Software
       matches.

    Methods
    -------
    __init__    The class constructor which initializes logging, and an empty
                dataframe for the resulting matches.

    load        Load CPE-SCCM software correspondance dataframe that was
                previously saved in pickled format.

    save        Save CPE-SCCM software correspondance dataframe to a
                serialized pickled flat file.

    get         Return a *copy* of the Vendor-Publisher matches dataframe.
    match       Prepare SCCM software data for ML Classification.
                Prepare CPE software data for ML Classification.
                Form the Cartesian product of potential software matches for
                each Vendor-Publisher tuple.
                Update the new data to be classified using manually labelled
                data. This ensures that only new unclassified data is input
                to the ML Matching.
                Run the ML Classification to match CPE Software to SCCM
                Software.
                Eliminate duplicate matches by choosing the "best" match
                based on a "Fuzzy Matching" statistic.
                Concatenate the new classified data with the manually
                labelled data to form the final result dataframe.

    Exceptions
    ----------
    IOError     Log error message and ignore

    Returns
    -------
    None

    """

    def __init__(self, mylogger=None):
        """Initialize class by configuring logging,  initializing dataframe.

        This is the class constructor. It initializes logging and allocates
        an empty dataframe to contain sccm hosts data.

        I/P Parameters
        --------------
        mylogger    logging object. If None, then a new object is initialzed.

        """
        # Configure logging

        if mylogger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(gbls.loglvl)
        else:
            self.logger = mylogger

        self.logger.info('\n\nInitializing MatchSoft class\n\n')

        self.df_match_cpe_sft = pd.DataFrame({
            # For each CPE Vendor_X, that vendor's software is mapped to
            # the corresponding SCCM software
            'vendor_X': [],
            'DisplayName0': [],
            'Version0': [],
            't_cve_name_x': [],
            't_cve_name_y': [],
            't_cve_name': []
            })

        # Initialize ML Classification
        self.__ML = ml.MLClassify('software')

        return None

    def load(self, mypck=None):
        """Load CPE-SCCM software match dataframe that was saved."""
        self.logger.info(
                '\n\nLoading CPE - SCCM software correspondance data into '
                'df_match_cpe_sft dataframe\n\n'
                )

        if mypck is None:
            mypck = gbls.df_match_cpe_sft_pck

        self.df_match_cpe_sft = pd.read_pickle(mypck)
        return None

    def save(self):
        """Save the CPE-SCCM software match dataframe."""
        self.logger.info('\n\nSaving df_match_cpe_sft dataframe\n\n')
        self.df_match_cpe_sft.to_pickle(gbls.df_match_cpe_sft_pck)
        return None

    def get(self):
        """Return a *copy* of the dataframe."""
        df_tmp = self.df_match_cpe_sft.copy()
        self.logger.info(
                '\n\nGet MatchVendor.df_match_cpe_sft: \n{0}\n{1}\n\n'.format(
                                df_tmp.shape,
                                df_tmp.columns
                                )
                )
        return df_tmp

    def match(
                self,
                df_match_vendor_publisher,
                df_sccm_ar,
                df_cpe4
                ):
        """Match the CPE "Software" data to the SCCM "Software" data.

        Actions
        -------

        *  The SCCM Software data is prepared for ML Classification:
                - Data is normalized.
                - Extraneous columns are dropped.
                - Missing values are initialized.
                - Vendor_Publisher correspondance dataframe is used to add
                  CPE Vendor to the SCCM record.
                - Data is grouped and sorted.

        *  Likewise the CPE Software data is prepared for ML Classification.
           Extraneous columns are dropped and data is grouped / sorted.

        *  The cartesian product is formed for each Vendor-Publisher tuple:
           {CPE "Software" for this vendor} X
                {SCCM Software installed for this "Publisher"}

            - Cisco Webex and all Microsoft software is ignored.
            - Java release data is normalized.
            - Simple heuristics are used to eliminate improbable matches.
            - "Fuzzy matching" statistics are calculated as ML "features".

        *  Manually labelled match data is read in. The new data to be
           classified is updated using this labelled data. This ensures that
           the ML Classification will run only on new unknown data.

        *  The Random Forest Classification algorithm is run.


        *  Duplicate matches are eliminated from both the newly classified
           data as well as the manually labelled data.

            - Only +ve matches are considered.
            - The row which has the highest fuzzy match statistic is
              picked as the "best" match.
            - The rest of the duplicate data in the group is discarded.

        *  The newly-classifed test data is concatenated with the manually
           labelled data to form the final dataframe of Vendor - Publisher
           correspondances.

        Exceptions
        ----------
        IOError     Log error message and ignore

        Returns
        -------
        None

        """
        ######
        # Utility Function to handle Java 'Release' data programmatically
        ######

        def _fix_java_rel(my_cpe_sft):

            # Java "release" data has varied widely over the years. The
            # transition from Sun to Oracle complicated things further.
            #
            # This rtn attempts to deal with the exception cases in order
            # to normalize the data.

            # handle cases where release is of form 'update[_]nn'

            if 'update' in my_cpe_sft:
                pattern = re.compile(
                                'cpe:2.3:a:(oracle|sun):'
                                '(?P<sft>(jdk|jre)):'
                                '1.(?P<rel>[\d\.]*):'
                                'update[_]*(?P<upd>[\d]*)',
                                re.IGNORECASE | re.UNICODE
                                )

                my_match = pattern.search(my_cpe_sft)

                if my_match:
                    sft = my_match.group('sft')
                    rel = my_match.group('rel')
                    upd = my_match.group('upd')

                    tmp_str = (
                                my_match.group('rel')
                                + '.'
                                + my_match.group('upd')
                                + '0'
                                )

                # if no match, then either not jre/jdk or version info missing

                else:
                    return '-'

            # the other possibilities is that only 'release' is provided
            else:
                pattern = re.compile(
                            'cpe:2.3:a:'
                            '(oracle|sun):'
                            '(?P<sft>(jdk|jre)):'
                            '1.(?P<rel>[\d\.\_]*):',
                            re.IGNORECASE | re.UNICODE
                            )

                my_match = pattern.search(my_cpe_sft)

                if my_match:
                    sft = my_match.group('sft')
                    rel = my_match.group('rel')

                    tmp_str = my_match.group('rel')

                # if no match, then either not jre/jdk or version info missing
                else:
                    return '-'

            # Handle case of a really old version: Has embedded '_'
            if '_' in rel:
                return ('1.' + rel)

            # JRE version is good 'as is'
            elif sft == 'jre':
                        return(tmp_str)

            # JDKS up to / including 1.7.x have release 1.7.x
            # JDKS 1.8.x have release '8.x'

            elif sft == 'jdk':

                if rel.startswith('8.'):
                    return tmp_str
                else:
                    return('1.' + tmp_str)
            else:
                self.logger.info('*** error rtn java')

        # Prepare the SCCM data for ML classification

        def _match_prepare_sccm_data(
                            df_add_rem_g,
                            df_match_vendor1
                            ):

            # Prepare sccm data for 2cd phase of machine learning matching
            # - Clean and normalize SCCM software data. Remove extraneous
            #   columns.
            # - Add CPE Vendor_X based on Publisher0-VendorX correspondance
            # - Group SCCM data by vendor_X, software name, release

            ######
            #   Access SCCM software inventory data
            ######

            self.logger.info('\n\nEntering match_prepare_sccm_data\n\n')

            # lower-case Publisher0 string for matching
            df_add_rem_g0 = df_add_rem_g
            df_add_rem_g0['publisher0'] = df_add_rem_g0[
                                                'Publisher0'
                                                ].str.lower()

            df_add_rem_g1 = df_add_rem_g0.drop('Publisher0', 1)

            self.logger.info(
                        '\n\nSCCM inventory dataframe \n{0}\n{1}\n\n'.format(
                            df_add_rem_g1.shape,
                            df_add_rem_g1.columns
                            )
                        )

            # Merge CPE vendor data with SCCM Publisher0 data from SCCM
            # inventory

            df_add_rem_g3 = pd.merge(
                                df_add_rem_g1,
                                df_match_vendor_publisher,
                                how='left',
                                on='publisher0'
                                )

            self.logger.info(
                        '\n\nMerged CPE vendor - '
                        'SCCM inventory dataframe \n{0}\n{1}\n\n'.format(
                            df_add_rem_g3.shape,
                            df_add_rem_g3.columns
                            )
                        )

            # Clean up the SCCM inventory data and prepare it for product
            # matching

            # Remove extraneous columns
            df_arSft = df_add_rem_g3[[
                                'vendor_X',
                                'DisplayName0',
                                'Version0'
                                ]]

            # and initialize missing release values. NVD CPE uses '-' for
            # missing release data, and so will we.

            df_arSft.loc[df_arSft.Version0.isnull(), 'Version0'] = '-'

            self.logger.info(
                        '\n\nSCCM Inventory dataframe summary:\n'
                        'Add_Remove Publishers:\n{0}\n'
                        'AR software:\n{1}\n\n'.format(
                            df_arSft['vendor_X'].nunique(),
                            df_arSft['DisplayName0'].nunique()
                            )
                        )

            # Group SCCM data by vendor_X, software name, release

            df_arSft_grp = df_arSft.groupby([
                                    'vendor_X',
                                    'DisplayName0',
                                    'Version0'
                                    ])

            return (df_arSft_grp)

        # Prepare the CPE Software data for ML classification

        def _match_prepare_cpe_data(df_cpe4):

            # Read in the CPE vendor / software data and prepare it for
            # product matching

            self.logger.info(
                        '\n\n Starting prepare_cpe_data\n'
                        'cpe data\n{0}\n{1}\n\n'.format(
                            df_cpe4.shape,
                            df_cpe4.columns
                            )
                        )

            # keep relevant columns
            df_cpeSoft = df_cpe4[[
                                'vendor_X',
                                'software_X',
                                'release_X',
                                'title_X',
                                'cpe23-item-name',
                                '@name'
                                ]]

            self.logger.debug(
                            'CPE vendors:\n{0}\n\n'.format(
                                df_cpeSoft['vendor_X'].nunique()
                                )
                            )

            self.logger.debug(
                            'CPE products:\n{0}\n\n'.format(
                                    df_cpeSoft['software_X'].nunique()
                                    )
                            )

            # Group CPE data by vendor_X

            df_cpeSoft_grp = df_cpeSoft.groupby(['vendor_X'])

            self.logger.info(
                        '\nCPE vendors: \n{0}\n'
                        'CPE products: \n{1}\n\n'
                        '{2}\n{3}\n'.format(
                                df_cpeSoft['vendor_X'].nunique(),
                                df_cpeSoft['software_X'].nunique(),
                                df_cpeSoft.shape,
                                df_cpeSoft.columns
                                )
                        )

            return (df_cpeSoft_grp)

        # Form the cartesian product is formed for each Vendor-Publisher
        # tuple:
        #       {CPE "Software" for this vendor} X
        #          {SCCM Software installed for this "Publisher"}

        # This gives a set of potential matches to input to ML
        # classification.

        def _cartesian_product(
                        df_arSft_grp,
                        df_cpeSoft_grp
                        ):

            # Loop thru the data to find potential matches

            self.logger.info('\n\nEntering cartesian_product\n\n')

            # list of product tuples to check
            lst_dict = []

            t0 = time()

            self.logger.info(
                        '\n\nStarting generation of '
                        'cartesian product of CPE products '
                        'to SCCM software ... \n'
                        '*** This can take some time - '
                        'maybe 5 min or more.\n\n'
                        )
            n = 0
            m = 0
            for key, df_ar_grp in df_arSft_grp:

                # split out vendor / SCCM DisplayName0 / SCCM Version0 strings
                (t_ar_vndrX, t_ar_dsply0, t_ar_ver0) = key

                # microsoft will be handled separately as service bulletins
                if t_ar_vndrX == 'microsoft':
                    continue

                # some cisco webex products are also hard to match
                if t_ar_vndrX == 'cisco':
                    if t_ar_ver0 == '-' and 'webex' in t_ar_dsply0.lower():
                        continue

                # get the corresponding CPE data for this vendor
                try:
                    df_cpe_grp = df_cpeSoft_grp.get_group((t_ar_vndrX))

                except KeyError as e:
                    self.logger.critical(
                        '\n\n***matchsft.py cartes product loop -'
                        ' KeyError: {0}\n\n'.format(e)
                        )
                    continue

                for (
                        t_cpeix,
                        t_cpe_vdr_X,
                        t_cpe_sft_X,
                        t_cpe_relX,
                        t_cpe_titleX,
                        t_cpe23_name,
                        t_cve_name
                        ) in df_cpe_grp.itertuples():

                    # 'normal' CPE release #
                    t_cpe_relX_tmp = t_cpe_relX

                    # but .... java - an exception (as always!)
                    if t_cpe_vdr_X in ['oracle', 'sun']:
                        if t_cpe_sft_X == 'jre' or t_cpe_sft_X == 'jdk':
                            t_cpe_relX_tmp = _fix_java_rel(t_cpe23_name)

                    # don't consider vendor name in fuzzy matching

                    t_cpe_titleX_tmp = t_cpe_titleX.lower().replace(
                                                                t_cpe_vdr_X,
                                                                ' '
                                                                )
                    t_ar_dsply0_tmp = t_ar_dsply0.lower().replace(
                                                            t_cpe_vdr_X,
                                                            ' '
                                                            )
                    ######
                    #   Apply quick heuristics to reduce the number of
                    #   possible matches
                    ######

                    # 1) Release #'s should at least partially match

                    fz_rel_ratio = fz.ratio(
                                        t_cpe_relX_tmp,
                                        t_ar_ver0
                                        )
                    fz_rel_ptl_ratio = fz.partial_ratio(
                                                t_cpe_relX_tmp,
                                                t_ar_ver0
                                                )

                    if (t_cpe_relX_tmp != '-') and (t_ar_ver0 != '-'):

                        # If release data is specified, then check that
                        # there is at least a partial match

                        if fz_rel_ratio < 90 or fz_rel_ptl_ratio < 100:
                            continue

                    # 2) There should be at least one occurence of one word in
                    # the cpe full name somewhere in sccm full name

                    fz_ptl_tok_set_ratio = fz.partial_token_set_ratio(
                                                t_cpe_titleX_tmp,
                                                t_ar_dsply0_tmp,
                                                force_ascii=False
                                                )

                    if fz_ptl_tok_set_ratio < 70:
                        continue

                    ######
                    # calculate fuzzy matching statistics for this match
                    ######

                    lst_dict.append({
                        'vendor_X': t_cpe_vdr_X,
                        'software_X': t_cpe_sft_X,
                        'Version0': t_ar_ver0,
                        'release_X': t_cpe_relX,
                        'title_X': t_cpe_titleX,
                        'DisplayName0': t_ar_dsply0,

                        'fz_ratio': fz.ratio(
                                        t_cpe_titleX_tmp,
                                        t_ar_dsply0_tmp
                                        ),
                        'fz_ptl_ratio': fz.partial_ratio(
                                                t_cpe_titleX_tmp,
                                                t_ar_dsply0_tmp
                                                ),
                        'fz_tok_set_ratio': fz_ptl_tok_set_ratio,
                        'fz_ptl_tok_sort_ratio': fz.token_sort_ratio(
                                                    t_cpe_titleX_tmp,
                                                    t_ar_dsply0_tmp,
                                                    force_ascii=False
                                                    ),
                        'fz_uwratio': fz.UWRatio(
                                        t_cpe_titleX_tmp,
                                        t_ar_dsply0_tmp
                                        ),

                        'fz_rel_ratio': fz_rel_ratio,
                        'fz_rel_ptl_ratio': fz_rel_ptl_ratio,
                        't_cve_name': t_cve_name
                        })
                    m = m+1

                n = n+1
                if n % 100 < 1:
                    self.logger.debug(
                            '---Working ar: '
                            'sccm sft i/p: {0} '
                            ', potential matches output: {1}\n'.format(n, m)
                            )
                # # debug code to speed thru looping process
                # if n > 2000:
                #     break

            duration = time() - t0
            self.logger.info(
                        '\n\nDone in {0} sec.\n\n'.format(
                                                duration
                                                )
                        )

            df_match = pd.DataFrame(lst_dict)

            if df_match.empty:
                self.logger.info(
                    '\n\nResulting cartesian product is empty\n\n'
                    )
                return (df_match)

            else:
                df_match1 = df_match[[
                                'vendor_X',
                                'software_X',
                                'title_X',
                                'DisplayName0',
                                'release_X',
                                'Version0',
                                'fz_ratio',
                                'fz_ptl_ratio',
                                'fz_tok_set_ratio',
                                'fz_ptl_tok_sort_ratio',
                                'fz_uwratio', 'fz_rel_ratio',
                                'fz_rel_ptl_ratio',
                                't_cve_name']]

                # add in length of names as features

                df_match1['titlX_len'] = df_match1['title_X'].apply(len)
                df_match1['DsplyNm0_len'] = df_match1[
                                                    'DisplayName0'
                                                    ].apply(len)

                self.logger.info(
                            '\n\n Results of matching: \n'
                            '# matches: {0}\n'
                            '# vendors: {1}\n'
                            '# CPE software: {2}\n'
                            '# SCCM inventory products: {3}\n'
                            '{4}\n{5}'.format(
                                        df_match1['t_cve_name'].count(),
                                        df_match1['vendor_X'].nunique(),
                                        df_match1['software_X'].nunique(),
                                        df_match1['DisplayName0'].nunique(),
                                        df_match1.shape,
                                        df_match1.columns
                                        )
                            )
                return (df_match1)

        # Previously manually labelled data is used to update the new data to
        # be classified. This ensures that only new, unknown data is input
        # into the ML Classification algorithm.

        def _update_with_labelled_data(df_match1):

            self.logger.info('\n\nEntering update_with_labelled_data\n\n')

            # Input the manually labelled data. This is a subset of the
            # 'match' data for selected vendors. This data was originally used
            # to train the classification model. We recycle it here since it
            # has already been classified (manually).

            try:
                labelled_sft_data = pd.io.parsers.read_csv(
                                                gbls.df_label_software,
                                                sep=gbls.SEP2,
                                                error_bad_lines=False,
                                                warn_bad_lines=True,
                                                index_col=0,
                                                encoding='utf-8')
            except IOError as e:
                self.logger.critical('\n\n***I/O error({0}): {1}\n\n'.format(
                            e.errno, e.strerror))

            except:
                self.logger.critical(
                    '\n\n***Unexpected error: {0}\n\n'.format(
                        sys.exc_info()[0]))
                raise

            self.logger.info(
                            '\n\nRead in manually matched '
                            'product-software '
                            'dataframe\n{0}\n{1}\n{2}\n\n'.format(
                                labelled_sft_data.shape,
                                labelled_sft_data.columns,
                                labelled_sft_data['match'].value_counts()
                                )
                            )

            # Use the manually labelled data to update the data to be ML
            #   matched.
            # Merge in 'match' information from labelled data

            df_match1_lbl = self.__ML.upd_using_labelled_data(
                                            df_match1,
                                            labelled_sft_data
                                            )
            self.logger.debug(
                '\n\nUpdated dframe\n'
                '{0}\n{1}\n\n'.format(
                    df_match1_lbl.shape,
                    df_match1_lbl.columns
                    )
                )

            # make sure the manually 'matched' data set is still intact

            if df_match1_lbl.empty:
                self.logger.info('Updated dframe is empty')
            else:
                self.logger.info(
                    '\n\nResults of updating the CPE software - '
                    'SCCM inventory dataframe with labelled data: \n'
                    '# matches: {0} \n'
                    '# vendors: {1} \n'
                    '# CPE software: {2} \n'
                    '# SCCM inventory products: {3} \n\n'.format(
                        df_match1_lbl['t_cve_name'].count(),
                        df_match1_lbl['vendor_X'].nunique(),
                        df_match1_lbl['software_X'].nunique(),
                        df_match1_lbl['DisplayName0'].nunique(),
                        df_match1_lbl.shape,
                        df_match1_lbl.columns
                        )
                )

            return (df_match1_lbl, labelled_sft_data)

        # Concatentate the manually labelled data with the newly classified
        # data. Drop extraneous columns and eliminate duplicate data to form
        # the final dataframe of CPE-SCCM "Software" matches.

        def _post_process_matched_data(
                            p_df_match1_test2,
                            p_labelled_sft_data
                            ):

            # Utility fn to eliminate duplicates

            # - Duplicates are grouped. Only +ve matches are considered.
            # - Then a simple heuristic is used to pick the "best" match.
            # - The rest of the duplicates are discarded.

            def replace_dups_by_best_match(p_df_match1_test2a):

                self.logger.info(
                    '\n\nEntering replace_dups_by_best_match\n\n'
                    )

                if p_df_match1_test2a.empty:
                    self.logger.critical(
                            '\n\n*** replace_dups - No hosts I/P!\n\n'
                            )
                    return p_df_match1_test2a

                # Force copy by value
                df_match1_test2a = p_df_match1_test2a.copy()

                self.logger.debug(
                        '\n\nDframe before removing duplicates '
                        '\n{0}\n{1}\n{2}\n\n'.format(
                            df_match1_test2a.shape,
                            df_match1_test2a.columns,
                            df_match1_test2a.apply(pd.Series.nunique)
                            )
                        )

                # Are only interested in +ve matches
                df_match1_test2a = df_match1_test2a[
                                        df_match1_test2a.match == 1
                                        ]

                # Pick out the row in group which has maximum fz_uwratio
                # value. This is likely to be the best match possible.

                df_match1_tst2a_gp = df_match1_test2a.groupby([
                                                            "vendor_X",
                                                            "DisplayName0",
                                                            "Version0"])

                df_match1_tst2b = df_match1_test2a[
                                        df_match1_test2a['fz_uwratio']
                                        == df_match1_tst2a_gp[
                                                'fz_uwratio'
                                                ].transform(max)]

                # Clean out the rest of the duplicates for the ML-classified
                # data
                df_match1_tst2c = df_match1_tst2b.drop_duplicates(
                                                        subset=[
                                                            u'vendor_X',
                                                            u'DisplayName0',
                                                            u'Version0']
                                                            )

                self.logger.info(
                        '\n\nDframe with duplicates removed '
                        '\n{0}\n{1}\n{2}\n\n'.format(
                            df_match1_tst2c.shape,
                            df_match1_tst2c.columns,
                            df_match1_tst2c.apply(pd.Series.nunique)
                            )
                        )
                return(df_match1_tst2c)

            ######
            # Main code for _post_process_matched_data function
            ######

            self.logger.info('\n\nEntering post_process_matched_data\n\n')

            # Force "call by value"

            df_match1_test2 = p_df_match1_test2.copy()
            labelled_sft_data = p_labelled_sft_data.copy()

            if df_match1_test2.empty:
                self.logger.critical(
                        '\n\n\n*** Product-software dframe is empty.\n\n')
                self.df_match_cpe_sft = df_match1_test2
            else:

                self.logger.info(
                            'Product-software match '
                            'dataframe\n{0}\n{1}\n{2}\n\n'.format(
                                df_match1_test2.shape,
                                df_match1_test2.columns,
                                df_match1_test2['match'].value_counts()
                                )
                            )

                # Eliminate duplicates from the ML Classified data

                df_match1_tst2d = replace_dups_by_best_match(df_match1_test2)

                self.logger.info(
                        '\n\nManually labelled '
                        'data\n {0}\n {1}\n {2}\n\n'.format(
                            labelled_sft_data.shape,
                            labelled_sft_data.columns,
                            labelled_sft_data['match'].value_counts()
                            )
                        )

                # Idem for the manually labelled data

                df_match1_lbl3 = replace_dups_by_best_match(
                                                labelled_sft_data)

                # Consolidate the new ML data with the manually labelled data

                df_match1_lbl4 = self.__ML.post_process_matched_data(
                                                    df_match1_tst2d,
                                                    df_match1_lbl3
                                                    )
                # drop extraneous columns
                self.df_match_cpe_sft = df_match1_lbl4[[
                                            'vendor_X',
                                            'DisplayName0',
                                            'Version0',
                                            't_cve_name'
                                            ]]

                self.logger.info(
                    '\n\nConsolidated product-software '
                    'match dataframe after removing '
                    'duplicates\n{0}\n{1}\n\n'.format(
                                    self.df_match_cpe_sft.shape,
                                    self.df_match_cpe_sft.columns
                                    )
                    )
            return None

        ######
        # Main code for "match" method
        ######

        self.logger.info('\n\nEntering match_soft\n\n')

        self.logger.info(
                        '\nSCCM Inventory dataframe shape: \n{0}\n\n'.format(
                                                df_sccm_ar.shape
                                                )
                        )

        # Prepare SCCM software data for matching

        df_arSft_grp = _match_prepare_sccm_data(
                                df_sccm_ar,
                                df_match_vendor_publisher
                                )

        # Prepare CPE vendor / software data for matching

        df_cpeSoft_grp = _match_prepare_cpe_data(df_cpe4)

        # Form cartesian product of CPE vendor/software data with SCCM
        # inventory software data

        df_match1 = _cartesian_product(
                        df_arSft_grp,
                        df_cpeSoft_grp
                        )

        # Update the set of potential match data with known matches
        # from the labelled data

        (df_match1_lbl, labelled_sft_data) = _update_with_labelled_data(
                                                                    df_match1
                                                                    )
        # Do actual ML classification

        (df_match1_test2, df_match_labelled) = self.__ML.ml_classify(
                                                            df_match1_lbl)

        # Post-process the ML matched data and the labelled data to
        # eliminate duplicates and then concatenate into one dataframe

        _post_process_matched_data(df_match1_test2, labelled_sft_data)

        return None



"""Initialize logging, global variables.

setup_logging: Initialize logging

init_globals: Initialize global variables

"""
import os
import json
import logging.config
import StringIO as strIO
import zipfile as zipf

import requests
from yapsy.PluginManager import PluginManager

import gbls

utils_logger = logging.getLogger(__name__)


def setup_logging(
        default_path=gbls.confdir + 'logging.json',
        default_level=logging.INFO,
        env_key='LOG_CFG'
        ):
    """Configure and initialize logging.

    I/P Parameters
    ==============

    default_path:   Directory / filename for json log output
    default_level:  Default logging level
    env_key:        Environment variable to configure logging.

    Return Value
    ============

    Returns None.

    """
    path = default_path
    value = os.getenv(env_key, None)
    if value:
        path = value
    if os.path.exists(path):
        with open(path, 'rt') as f:
            config = json.load(f)
        logging.config.dictConfig(config)
    else:
        logging.basicConfig(level=default_level)

    # Initialize logging for utils itself

    utils_logger.setLevel(gbls.loglvl)

    utils_logger.info('\n\nUtils: Logging initialized\n\n')

    return None


def init_globals():
    """Initialize global variables."""
    gbls.pckdir = gbls.wkdir + 'pck/'
    gbls.confdir = gbls.wkdir + 'conf/'
    gbls.csvdir = gbls.wkdir + 'csv/'
    gbls.nvddir = gbls.wkdir + 'nvd/'
    gbls.modeldir = gbls.wkdir + 'models/'

    # Create directories if do not exist
    if not os.path.isdir(gbls.pckdir):
        os.makedirs(gbls.pckdir)
    if not os.path.isdir(gbls.nvddir):
        os.makedirs(gbls.nvddir)

    ######
    #   CSV Input data
    ######

    gbls.v_r_system = gbls.csvdir + 'v_R_System.csv'
    gbls.v_gs_add_rem_pgms = gbls.csvdir + 'v_GS_ADD_REMOVE_PROGRAMS.csv'
    gbls.v_gs_add_rem_pgms_64 = (
                            gbls.csvdir
                            + 'v_GS_ADD_REMOVE_PROGRAMS_64.csv'
                            )
    gbls.ad_vip_grps = gbls.csvdir + 'ps-ad-vip.csv'
    gbls.s_vndr_stop_wds = gbls.csvdir + 's_vndr_stop_wds.csv'
    gbls.df_label_software = gbls.csvdir + 'label_software.csv'
    gbls.df_label_vendors = gbls.csvdir + 'label_vendors.csv'

    ######
    #   Saved dataframe names
    #   'rf_' - refactor code version
    ######
    gbls.df_sys_pck = gbls.pckdir + 'rf_df_sys.pck'
    gbls.df_add_rem_g_pck = gbls.pckdir + 'rf_df_add_rem_g.pck'
    gbls.df_cpe4_pck = gbls.pckdir + 'rf_df_cpe4.pck'
    gbls.df_cve_pck = gbls.pckdir + 'rf_df_cve.pck'
    gbls.df_v_R_System_3modified_pck = (
                        gbls.pckdir
                        + 'rf_df_v_R_System_3modified.pck'
                        )
    gbls.df_sft_vuln_pck = gbls.pckdir + 'rf_gbls.df_sft_vuln.pck'
    gbls.df_match_vendor_publisher_pck = (
                        gbls.pckdir
                        + 'rf_df_match_vendor_publisher.pck'
                        )
    gbls.df_match_cpe_sft_pck = (
                        gbls.pckdir
                        + 'rf_df_match_cpe_sft.pck'
                        )

    ######
    #   NVD data
    ######

    # https://static.nvd.nist.gov/feeds/xml/cve/2.0/nvdcve-2.0-2016.meta
    gbls.url_meta_base = (
                    'https://static.nvd.nist.gov'
                    + '/feeds/xml/cve/2.0/nvdcve-2.0-'
                    )
    gbls.url_meta_end = '.meta'

    # https://nvd.nist.gov/feeds/xml/cve/nvdcve-2.0-2016.xml.zip
    gbls.url_xml_base = 'https://nvd.nist.gov/feeds/xml/cve/nvdcve-2.0-'
    gbls.url_xml_end = '.xml.zip'

    gbls.url_cpe = (
                'http://static.nvd.nist.gov/feeds/xml/'
                'cpe/dictionary/official-cpe-dictionary_v2.3.xml.zip'
                )

    # Hardcode the cpe filename in case.
    gbls.cpe_filename = 'official-cpe-dictionary_v2.3.xml'
    gbls.cve_filename = 'nvdcve-2.0-'

    # Update the cpe data if older than this value (in days)
    gbls.cpe_max_age = 7

    gbls.nvd_meta_filename = 'my_meta_'

    # The cpe dictionary lists standardized vendor - software names
    gbls.nvdcpe = gbls.nvddir + gbls.cpe_filename

    # The cve vulnerability data files are named by year
    #    eg nvdcve-2.0-2015.xml

    gbls.nvdcve = gbls.nvddir + gbls.cve_filename

    ######
    #   ML Models
    ######
    gbls.clf_vendor = (
                gbls.modeldir +
                'vendor_classif_trained_Rdm_Forest.pkl.z'
                )
    gbls.clf_software = (
                gbls.modeldir +
                'software_classif_trained_Rdm_Forest.pkl.z'
                )

    return None

def load_plugins():
    """Load plugins.

    I/P Parameters
    ==============

    None.

    Actions
    =======

    1. Sets yapsy logger's logging level to the global default.
    2. Loads the (one and only) plugin from the plugin directory.
    3. Invokes the plugin's "print_name" method to print the name.

    Return Value
    ============

    Returns None.

    """
    utils_logger.info('\n\nEntering load_plugins\n\n')

    # Set logging for the yapsy plugin framework
    logging.getLogger('yapsy').setLevel(gbls.loglvl)

    # Load the plugins from the plugin directory.
    gbls.plugin_manager = PluginManager()
    gbls.plugin_manager.setPluginPlaces([gbls.PLUGINFOLDER])
    gbls.plugin_manager.collectPlugins()

    # Loop through the plugins and print their names.
    for plugin in gbls.plugin_manager.getAllPlugins():
        plugin.plugin_object.print_name()

    return None

def get_zip(myurl):
    """Download and unzip a file

    This utility rtn downloads a file given the URL and then unzips it.

    Parameters
    ==========

    myurl   The URL of the file to be downloaded

    Returns
    =======
    (file_name, extracted_file)
            Filename    Name in file in the zip archive
            extracted_file
                        Contents of file

            (None, None) is returned if an error is detected.

    Exceptions
    ==========
    RequestException:   The requests module, used for https access, has
                    several exception conditions.

    Restrictions
    ============

    This rtn is designed to be used for NIST XML file downloads.
    The assumption is that the archive contains only 1 XML in zipped format.
    The contents of the file are read into memory.

    """
    utils_logger.info(
        '\n\nEntering get_zip to read {0}\n\n'.format(
                                            myurl
                                            )
        )

    try:
        resp = requests.get(myurl)

    except requests.exceptions.RequestException as e:
            utils_logger.critical(
                '\n\n***NVD XML feeds - Error: \n{0}\n{0}\n\n'.format(
                    myurl,
                    e
                    )
                )
            return (None, None)

    # unzip compressed archive
    my_zipfile = zipf.ZipFile(strIO.StringIO(resp.content))
    zip_names = my_zipfile.namelist()

    # should be only 1 file in the archive
    if len(zip_names) == 1:
        file_name = zip_names.pop()
        extracted_file = my_zipfile.open(file_name).read()
        utils_logger.info(
            'get_zip: Successfully extracted {0}'.format(
                                                        file_name
                                                        )
            )
        return (file_name, extracted_file)
    else:
        utils_logger.critical(
            'get_zip: Error in extracting NVD zip file'
            )
        return (None, None)



"""Input / Process SCCM host and software data.

Purpose
=======

The sccm module inputs, parses, persists SCCM host and software data.

Public classes
==============

SccmHosts       SCCM Hosts class
SccmSoft        SCCM Software class

"""
import pandas as pd
import sys
import logging

import gbls
import utils

# Public classes
__all__ = (
        'SccmHosts',
        'SccmSoft'
        )


class SccmHosts(object):
    """Input, parse, persist SCCM-managed host data.

    Methods
    -------
    __init__    Class constructor to configure logging, initialize empty data
                frame
    read        Input the raw SCCM CSV file. Clean data. Remove columns.
    load        Load hosts dataframe from the pickled file.
    save        Save the hosts dataframe to the corresponding pickled file.
    get         Return a *copy* of the hosts dataframe.

    Restrictions
    ------------
    The class instance variable self.df_sys should not be accessed directly
    by external callers.

    """

    def __init__(self, mylogger=None):
        """Initialize class by configuring logging,  initializing dataframe.

        This is the class constructor. It initializes logging and allocates
        an empty dataframe to contain sccm hosts data.

        I/P Parameters
        --------------
        mylogger    logging object. If None, then a new object is initialzed.

        """
        # Configure logging

        if mylogger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(gbls.loglvl)
        else:
            self.logger = mylogger

        self.logger.debug('\n\nInitializing SccmHosts class\n\n')

        self.df_sys = pd.DataFrame({
            'ResourceID': []
            })

    def read(self, mydir=None):
        """Read the raw SCCM CSV host data.

        Actions
        =======

        Reads v_R_System view raw data. v_R_System is the SCCM view
        listing managed hosts.

        Cleans data:
            * Remove extraneous columns
            * Only keep valid entries; remove unmanaged / inactive hosts.

        Exceptions
        ==========
        IOError     Produce a message, then ignore.
        """
        # Read the i/p raw CSV file containing extracted SCCM software data

        self.logger.info('\n\nEntering SccmHosts.read_csv()\n\n')

        if mydir is None:
            mydir = gbls.v_r_system

        self.logger.debug(
                        'Reading file {0},\nsep: {1}\n\n'.format(
                                mydir,
                                gbls.SEP)
                        )

        try:
            df_sys_tmp = pd.io.parsers.read_csv(
                                    mydir,
                                    sep=gbls.SEP,
                                    error_bad_lines=False,
                                    warn_bad_lines=True,
                                    quotechar='"',
                                    encoding='utf-16')
            # Note the use of utf-16 for this data.
        except IOError as e:
            self.logger.critical('\n\n***I/O error({0}): {1}\n\n'.format(
                        e.errno, e.strerror))

        # except ValueError:
        #    self.logger.critical('Could not convert data to an integer.')
        except:
            self.logger.critical(
                '\n\n***Unexpected error: {0}\n\n'.format(
                    sys.exc_info()[0]))
            raise

        # Only keep interesting columns

        self.df_sys = df_sys_tmp.loc[
                            :, [
                                'ResourceID',
                                'Active0',
                                'AD_Site_Name0',
                                'Distinguished_Name0',
                                'Resource_Domain_OR_Workgr0'
                                ]]

        # Convert site data to a pandas category
        self.df_sys['Site_X'] = self.df_sys[
                                    'AD_Site_Name0'
                                    ].astype('category')

        # Remove inactive hosts
        self.df_sys = self.df_sys[self.df_sys.Active0 > 0]

        # self.logger.debug() basic information
        self.logger.debug('\nv_R_System: \n{0}\n{1}\n\n'.format(
            self.df_sys.shape,
            self.df_sys.columns))
        return None

    def load(self, mydir=None):
        """Load hosts dataframe that was previously pickled."""
        self.logger.info(
            '\n\nLoading v_R_System into '
            'SccmHosts.df_sys dataframe\n\n'
            )

        if mydir is None:
            mydir = gbls.df_sys_pck

        self.df_sys = pd.read_pickle(mydir)
        return None

    def save(self):
        """Save hosts dataframe in a serialized pickle flat file."""
        self.logger.info('\n\nSaving SccmHosts.df_sys dataframe\n\n')
        self.df_sys.to_pickle(gbls.df_sys_pck)
        return None

    def get(self):
        """Return a *copy* of the main hosts dataframe."""
        df_tmp = self.df_sys.copy()
        self.logger.info(
                '\n\n Get SccmHosts.df_sys: \n{0}\n{1}\n\n'.format(
                                df_tmp.shape,
                                df_tmp.columns
                                )
                )
        return df_tmp


class SccmSoft(object):
    """Read, process, persist SCCM Software inventory data.

    Actions
    =======
    The data extracted from the SCCM views v_gs_add_rem_pgms and
    v_gs_add_rem_pgms_64 is input and processed into a dataframe.

    Methods
    =======
    __init__    Class constructor to configure logging, initialize empty data
                frame
    read        Input the raw SCCM CSV file. Clean data and save in dataframe.
    load        Load hosts dataframe from the pickled file.
    save        Save the hosts dataframe to the corresponding pickled file.
    get         Return a *copy* of the hosts dataframe.

    """

    def __init__(self, mylogger=None):
        """Initialize class by configuring logging,  initializing dataframe.

        This is the class constructor. It initializes logging and allocates
        an empty dataframe to contain sccm software data.

        I/P Parameters
        --------------
        mylogger    logging object. If None, then a new object is initialzed.

        """
        # Configure logging

        if mylogger is None:
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(gbls.loglvl)
        else:
            self.logger = mylogger

        self.logger.debug('\n\nInitializing SccmSoft Class\n\n')

        # Empty dataframe

        self.df_add_rem_g_empty = pd.DataFrame({
            #   Fields in SCCM software record
            'ResourceID': [],
            'AgentID': [],
            'DisplayName0': [],
            'GroupID': [],
            'InstallDate0': [],
            'ProdID0': [],
            'Publisher0': [],
            'RevisionID': [],
            'TimeStamp': [],
            'Version0': [],

            #   Fields added during processing
            'arch_X': []
            })

        # Initialize to empty dataframe
        self.df_add_rem_g = self.df_add_rem_g_empty

    def read(self, mydir_x86=None, mydir_x64=None):
        """Read the raw SCCM CSV host data.

        Actions
        =======

        Reads v_GS_ADD_REMOVE_PROGRAMS and  v_GS_ADD_REMOVE_PROGRAMS_64 view
        raw data. These SCCM view list software installed on each managed
        host.

        The two tables are consolidated into one.

        A field is added to show the processor architecture.

        Entries with missing Publisher0 or DisplayName0 values are dropped
        since these records are not usable in any case.

        Microsoft inventory data is also dropped since MS software inventory /
        vuln information can be handled more efficiently using MS-specific
        files.

        Exceptions
        ==========
        IOError     Produce a message, then ignore.

        """
        self.logger.info('\n\nEntering SccmSoft.read()\n\n')

        if mydir_x86 is None:
            mydir_x86 = gbls.v_gs_add_rem_pgms

        if mydir_x64 is None:
            mydir_x64 = gbls.v_gs_add_rem_pgms_64


        self.logger.debug(
                            'files {0} \n   {1} \n   sep: {2}'.format(
                                mydir_x86,
                                mydir_x64,
                                gbls.SEP
                                )
                            )

        try:
            df_v_gs_add_rem_tmp = pd.io.parsers.read_csv(
                    mydir_x86,
                    sep=gbls.SEP,
                    error_bad_lines=False,
                    warn_bad_lines=True,
                    quotechar='"',
                    encoding='utf-16')

        except IOError as e:
            self.logger.critical('\n\n***I/O error({0}): {1}\n\n'.format(
                        e.errno, e.strerror))

        # ValueError Exception could mean empty data set read
        # Initialize an empty dataframe
        except ValueError as e:
            self.logger.critical(
                            '\n\n***Value error: {0}\n- empty data set '
                            'returned\n\n'.format(
                                        sys.exc_info()[0]
                                        )
                            )
            df_v_gs_add_rem_tmp = self.df_add_rem_g_empty

        except:
            self.logger.critical(
                '\n\n***Unexpected error: {0}\n\n'.format(
                    sys.exc_info()[0]))
            raise

        self.logger.debug(
                    '\n\nFinished reading x86 view, starting x64 view\n\n')

        try:
            df_v_gs_add_rem64_tmp = pd.io.parsers.read_csv(
                    mydir_x64,
                    sep=gbls.SEP,
                    error_bad_lines=False,
                    warn_bad_lines=True,
                    quotechar='"',
                    encoding='utf-16')

        except IOError as e:
            self.logger.critical('\n\n***I/O error({0}): {1}\n\n'.format(
                        e.errno, e.strerror))

        # ValueError Exception could mean empty data set read
        # Initialize an empty dataframe
        except ValueError as e:
            self.logger.critical(
                            '\n\n***Value error: {0}\n- empty data set '
                            'returned\n\n'.format(
                                        sys.exc_info()[0]
                                        )
                            )
            df_v_gs_add_rem64_tmp = self.df_add_rem_g_empty

        except:
            self.logger.critical(
                '\n\n***Unexpected error: {0}\n\n'.format(
                    sys.exc_info()[0]))
            raise

        self.logger.debug(
            '\n\nFinished reading both v_gs_add_rem_pgms_xx views\n\n')

        # indicate processor architecture
        df_v_gs_add_rem_tmp['arch_X'] = False
        df_v_gs_add_rem64_tmp['arch_X'] = True

        # consolidate the "GS_add_remove" dataframes
        df_add_rem_g0 = pd.concat(
                [df_v_gs_add_rem_tmp, df_v_gs_add_rem64_tmp],
                axis=0,
                join='outer'
                )

        # Print basic information
        self.logger.debug('\nv_gs_Add_Remove_Programs:\n{0}\n{1}\n\n'.format(
                df_add_rem_g0.shape,
                df_add_rem_g0.columns)
                )

        # drop rows with either missing Publisher0 or DisplayName0 data
        df_add_rem_g1 = df_add_rem_g0[
                    df_add_rem_g0.Publisher0.notnull()
                    & df_add_rem_g0.DisplayName0.notnull()
                    ]

        self.logger.debug(
            '\nSCCM inventory data-no missing '
            'data \n{0}\n{1}\n\n'.format(
                    df_add_rem_g1.shape,
                    df_add_rem_g1.columns
                    )
            )

        # dont want MS inventory data either
        self.df_add_rem_g = df_add_rem_g1[
                            ~df_add_rem_g1.Publisher0.str.contains(
                                    'microsoft',
                                    case=False)
                                    ]

        self.logger.info(
            '\n\nSCCM inventory data after removing '
            'entries with missing values and also '
            'microsoft-related entries \n{0}\n{1}\n\n'.format(
                    self.df_add_rem_g.shape,
                    self.df_add_rem_g.columns
                    )
            )

        return None

    def load(self, mydir=None):
        """Load Software dataframe that was previously pickled."""
        self.logger.info(
            '\n\nLoading v_gs_Add_Remove_Programs '
            'into SccmSoft.df_add_rem_g dataframe\n\n'
            )

        if mydir is None:
            mydir = gbls.df_add_rem_g_pck

        self.df_add_rem_g = pd.read_pickle(mydir)

        self.logger.info(
            '\n\nSCCM inventory data loaded: '
            '\n{0}\n{1}\n\n'.format(
                    self.df_add_rem_g.shape,
                    self.df_add_rem_g.columns
                    )
            )
        return None

    def save(self):
        """Save Software dataframe in serialized pickle format."""
        self.logger.info(
            '\n\nSaving SccmSoft.df_add_rem_g dataframe\n\n'
            )
        self.df_add_rem_g.to_pickle(
            gbls.df_add_rem_g_pck
            )
        return None

    def get(self):
        """Return a *copy* of the main software dataframe."""
        df_tmp = self.df_add_rem_g.copy()
        self.logger.info(
                '\n\nGet SccmHosts.df_add_rem_g: \n{0}\n{1}\n\n'.format(
                                df_tmp.shape,
                                df_tmp.columns
                                )
                )
        return df_tmp
